# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from hamcrest import assert_that, equal_to, none

__author__ = 'Cedric Zhuang'


def test_create_delete_iscsi_ip(vnx_gf):
    sp, port_id = vnx_gf.select_port()

    port = vnx_gf.vnx.get_iscsi_port(sp, port_id)[0]

    # create ip
    port1 = port.config_ip('5.5.5.5', '255.255.255.0', '5.5.5.1', 1, 1)
    vnx_gf.until_existed(port1)

    assert_that(port1.ip_address, equal_to('5.5.5.5'))
    assert_that(port1.subnet_mask, equal_to('255.255.255.0'))
    assert_that(port1.gateway_address, equal_to('5.5.5.1'))

    # delete ip
    port1.delete_ip()
    vnx_gf.until_not_existed(port1)

    port = vnx_gf.vnx.get_iscsi_port(sp, port_id)[0]
    assert_that(port.ip_address, none())
    assert_that(port.subnet_mask, none())
    assert_that(port.gateway_address, none())



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals, division

from unittest import TestCase

from hamcrest import assert_that, has_items, equal_to, raises, has_item, \
    close_to, is_not

from storops.unity import calculator
from storops.unity.calculator import calculators, IdValues, \
    delta_ps, mb_ps_by_block, busy_idle_util, \
    sp_delta_ps, sp_mb_ps_by_byte, sp_mb_ps_by_block, sp_busy_idle_util, \
    only_one_path, sp_fact, all_not_none, UnityMetricConfig, \
    UnityMetricConfigParser, total_delta_ps, sp_total_delta_ps, \
    system_delta_ps, system_total_delta_ps, disk_response_time, \
    disk_queue_length, lun_response_time, lun_queue_length, \
    sp_sum_values, sp_io_rate, byte_rate, total_byte_rate, sp_byte_rate, \
    sp_total_byte_rate, system_byte_rate, system_total_byte_rate
from storops.unity.resource.disk import UnityDisk
from storops.unity.resource.filesystem import UnityFileSystem
from storops_test.unity.resource.test_metric import qr_6, qr_14, qr_17, \
    qr_34, qr_128, qr_130
from storops_test.unity.rest_mock import patch_rest
from storops_test.utils import is_nan

__author__ = 'Cedric Zhuang'


class MockCli(object):
    def __init__(self, prev, curr):
        self.prev_counter = prev
        self.curr_counter = curr


class CalculatorMetaInfoTest(TestCase):
    def test_get_metric_names(self):
        names = calculators.get_metric_names(UnityDisk)
        assert_that(names, has_items(
            'read_iops', 'write_iops', 'read_mbps', 'write_mbps'))

    def test_all_paths_default(self):
        paths = calculators.get_all_paths()
        assert_that(paths, has_items(
            'sp.*.physical.disk.*.reads',
            'sp.*.physical.disk.*.writes',
            'sp.*.physical.disk.*.readBlocks',
            'sp.*.physical.disk.*.writeBlocks',
            'sp.*.cpu.summary.busyTicks',
            'sp.*.storage.filesystem.*.reads',
            'sp.*.storage.lun.*.reads'))

    def test_all_paths_with_clz_list(self):
        paths = calculators.get_all_paths([UnityDisk, UnityFileSystem])
        assert_that(paths, has_items(
            'sp.*.physical.disk.*.reads',
            'sp.*.storage.filesystem.*.reads'))
        assert_that(paths, is_not(has_item('sp.*.cpu.summary.busyTicks')))
        assert_that(paths, is_not(has_item('sp.*.storage.lun.*.reads')))

    def test_get_calculator_of_metric(self):
        calc = calculators.get_calculator('UnityDisk', 'read_iops')
        assert_that(calc, equal_to(delta_ps))

    @patch_rest
    def test_get_metric_value_normal(self):
        disk_counters = MockCli(qr_6, qr_14)
        value = calculators.get_metric_value(
            UnityDisk, 'read_iops', disk_counters, 'dae_0_1_disk_2')
        expected = ((4158667 + 5) - (2966780 + 5)) / 163800.0
        assert_that(value, equal_to(expected))

    @patch_rest
    def test_get_metric_value_no_prev_data(self):
        disk_counters = MockCli(None, qr_6)
        value = calculators.get_metric_value(
            UnityDisk, 'read_iops', disk_counters, 'dae_0_1_disk_2')
        assert_that(value, is_nan())


class IdValuesTest(TestCase):
    def setUp(self):
        self.o1 = IdValues({'a': 2, 'b': 3})
        self.o2 = IdValues({'a': 7, 'c': 13})
        self.o3 = IdValues({'a': None, 'b': 17})

    def test_add_none(self):
        o1 = self.o1
        r = o1 + None
        assert_that(o1['a'], equal_to(2))
        assert_that(o1['b'], equal_to(3))
        assert_that(r['a'], equal_to(2))
        assert_that(r['b'], equal_to(3))
        r.set('a', 5)
        assert_that(o1['a'], equal_to(2))
        assert_that(r['a'], equal_to(5))

    def test_sum(self):
        r = sum([self.o1, self.o2, self.o3])
        assert_that(r['a'], equal_to(9))
        assert_that(r['b'], equal_to(20))
        assert_that(r['c'], equal_to(13))

    def test_self_add(self):
        r = self.o1.copy()
        r += self.o2
        assert_that(r['a'], equal_to(9))
        assert_that(r['b'], equal_to(3))
        assert_that(r['c'], equal_to(13))

    def test_add_has_none(self):
        r = self.o1 + self.o3
        assert_that(r['a'], equal_to(2))
        assert_that(r['b'], equal_to(20))

    def test_radd_none(self):
        o1 = self.o1
        r = None + o1
        assert_that(o1['a'], equal_to(2))
        assert_that(o1['b'], equal_to(3))
        assert_that(r['a'], equal_to(2))
        assert_that(r['b'], equal_to(3))
        r.set('a', 5)
        assert_that(o1['a'], equal_to(2))
        assert_that(r['a'], equal_to(5))

    def test_add_id_value(self):
        o1 = self.o1
        o2 = self.o2
        r = o1 + o2
        assert_that(o1['a'], equal_to(2))
        assert_that(o1['b'], equal_to(3))
        assert_that(len(o1), equal_to(2))
        assert_that(r['a'], equal_to(9))
        assert_that(r['b'], equal_to(3))
        assert_that(r['c'], equal_to(13))
        assert_that(len(r), equal_to(3))

    def test_add_int(self):
        o1 = self.o1
        r = o1 + 5
        assert_that(o1['a'], equal_to(2))
        assert_that(o1['b'], equal_to(3))
        assert_that(len(o1), equal_to(2))
        assert_that(r['a'], equal_to(7))
        assert_that(r['b'], equal_to(8))
        assert_that(len(r), equal_to(2))

    def test_radd_float(self):
        o1 = self.o1
        r = 5.1 + o1
        assert_that(o1['a'], equal_to(2))
        assert_that(o1['b'], equal_to(3))
        assert_that(len(o1), equal_to(2))
        assert_that(r['a'], equal_to(7.1))
        assert_that(r['b'], equal_to(8.1))
        assert_that(len(r), equal_to(2))

    def test_sub_id_values(self):
        r = self.o2 - self.o1
        assert_that(self.o2['a'], equal_to(7))
        assert_that(self.o2['c'], equal_to(13))
        assert_that(r['a'], equal_to(5))
        assert_that(r['b'], equal_to(-3))
        assert_that(r['c'], equal_to(13))

    def test_rsub_none(self):
        r = None - self.o1
        assert_that(r['a'], equal_to(-2))
        assert_that(r['b'], equal_to(-3))

    def test_rsub_float(self):
        r = 9.5 - self.o1
        assert_that(r['a'], equal_to(7.5))
        assert_that(r['b'], equal_to(6.5))

    def test_add_from_empty(self):
        r = IdValues() + self.o1
        assert_that(r['a'], equal_to(2))
        assert_that(r['b'], equal_to(3))

    def test_div_id_value(self):
        r = self.o1 / self.o2
        assert_that(r['a'], equal_to(2.0 / 7.0))
        assert_that(r['b'], is_nan())
        assert_that(r['c'], is_nan())

    def test_div_numeric(self):
        r = self.o1 / 2
        assert_that(r['a'], equal_to(1.0))
        assert_that(r['b'], equal_to(1.5))

    def test_div_by_zero(self):
        r = self.o1 / 0
        assert_that(r['a'], is_nan())
        assert_that(r['b'], is_nan())

    def test_div_zero_dev(self):
        r = 0.0 / IdValues({'a': 0, 'b': 3})
        assert_that(r['a'], equal_to(0.0))
        assert_that(r['b'], equal_to(0.0))

    def test_div_by_partial_zero(self):
        r = self.o1 / IdValues({'a': 2, 'b': 0})
        assert_that(r['a'], equal_to(1.0))
        assert_that(r['b'], is_nan())

    def test_rdiv_numerical(self):
        r = 12 / self.o1
        assert_that(r['a'], equal_to(6))
        assert_that(r['b'], equal_to(4))

    def test_rdiv_none(self):
        r = None / self.o1
        assert_that(r['a'], is_nan())
        assert_that(r['b'], is_nan())

    def test_mul_none(self):
        r = self.o1 * None
        assert_that(r['a'], equal_to(2))
        assert_that(r['b'], equal_to(3))

    def test_rmul_none(self):
        r = None * self.o1
        assert_that(r['a'], equal_to(2))
        assert_that(r['b'], equal_to(3))

    def test_mul_numerical(self):
        r = self.o1 * 3
        assert_that(r['a'], equal_to(6))
        assert_that(r['b'], equal_to(9))

    def test_rmul_numerical(self):
        r = 3 * self.o1
        assert_that(r['a'], equal_to(6))
        assert_that(r['b'], equal_to(9))

    def test_mul_id_values(self):
        r = self.o1 * self.o2
        assert_that(r['a'], equal_to(14))
        assert_that(r['b'], equal_to(3))
        assert_that(r['c'], equal_to(13))

    def test_mul_partial_none(self):
        r = self.o1 * self.o3
        assert_that(r['a'], equal_to(2))
        assert_that(r['b'], equal_to(51))


class CalculatorTest(TestCase):
    @patch_rest
    def test_disk_read_iops(self):
        ret = delta_ps('sp.*.physical.disk.*.reads', qr_6, qr_14)
        expected = ((4158667 + 5) - (2966780 + 5)) / 163800.0
        assert_that(ret['dae_0_1_disk_2'], equal_to(expected))

        assert_that(delta_ps('sp.*.physical.disk.*.reads', qr_6, qr_14,
                             'dae_0_1_disk_2'),
                    equal_to(expected))

    @patch_rest
    def test_disk_read_obj_not_found(self):
        ret = delta_ps('sp.*.physical.disk.*.reads', qr_6, qr_14,
                       'not_found')
        assert_that(ret, is_nan())

    @patch_rest
    def test_disk_write_iops(self):
        ret = delta_ps('sp.*.physical.disk.*.writes', qr_6, qr_14)
        expected = ((3762339 + 5) - (2679611 + 5)) / 163800.0
        assert_that(ret['dae_0_1_disk_2'], equal_to(expected))

    @patch_rest
    def test_path_not_found(self):
        ret = mb_ps_by_block('sp.*.physical.disk.*.writeBlocks', qr_6, qr_14)
        assert_that(ret['dae_0_1_disk_2'], is_nan())

        ret = mb_ps_by_block(
            'sp.*.physical.disk.*.writeBlocks', qr_6, qr_14, 'dae_0_1_disk_2')
        assert_that(ret, is_nan())

    @patch_rest
    def test_mb_ps_by_block_value(self):
        ret = mb_ps_by_block('sp.*.physical.disk.*.writeBlocks', qr_17, qr_34)
        assert_that(ret['dae_0_1_disk_0'], equal_to(5.0))

    def test_delta_ps_error_path(self):
        def f():
            delta_ps(['a', 'b'], None, None)

        assert_that(f, raises(ValueError, 'only one path'))

    @patch_rest
    def test_disk_utilization_all(self):
        paths = ['sp.*.physical.disk.*.busyTicks',
                 'sp.*.physical.disk.*.idleTicks']
        ret = busy_idle_util(paths, qr_17, qr_34)
        expected = self.expected_disk_utilization()
        assert_that(ret['dae_0_1_disk_0'], equal_to(expected))
        assert_that(ret['dpe_disk_24'], is_nan())

    @patch_rest
    def test_disk_utilization_single(self):
        paths = ['sp.*.physical.disk.*.busyTicks',
                 'sp.*.physical.disk.*.idleTicks']
        ret = busy_idle_util(paths, qr_17, qr_34, 'dae_0_1_disk_0')
        expected = self.expected_disk_utilization()
        assert_that(ret, equal_to(expected))

    @patch_rest
    def test_disk_utilization_nan_single(self):
        paths = ['sp.*.physical.disk.*.busyTicks',
                 'sp.*.physical.disk.*.idleTicks']
        ret = busy_idle_util(paths, qr_17, qr_34, 'dpe_disk_24')
        assert_that(ret, is_nan())

    def expected_disk_utilization(self):
        busy = (55249062767 + 631119255518) - (55243132789 + 630939312447)
        idle = (609018780542 + 0) - (608836882713 + 0)
        return busy * 100.0 / (idle + busy)

    @patch_rest
    def test_sp_delta_ps_all(self):
        ret = sp_delta_ps('sp.*.cifs.smb1.basic.writes', qr_17, qr_34)
        assert_that(ret['spa'], close_to(2.77, 0.01))
        assert_that(ret['spb'], close_to(5.55, 0.01))

    @patch_rest
    def test_sp_delta_ps_obj(self):
        ret = sp_delta_ps('sp.*.cifs.smb1.basic.writes', qr_17, qr_34, 'spa')
        assert_that(ret, close_to(2.77, 0.01))

        ret = sp_delta_ps('sp.*.cifs.smb1.basic.writes', qr_17, qr_34, 'spc')
        assert_that(ret, is_nan())

    @patch_rest
    def test_sp_mb_ps_by_byte_all(self):
        ret = sp_mb_ps_by_byte('sp.*.cifs.smb1.basic.writeBytes', qr_17, qr_34)
        assert_that(ret['spa'], equal_to(1.0))
        assert_that(ret['spb'], equal_to(2.0))

    @patch_rest
    def test_sp_mb_ps_by_block_all(self):
        ret = sp_mb_ps_by_block(
            'sp.*.storage.summary.writeBlocks', qr_17, qr_34)
        assert_that(ret['spa'], equal_to(30.0))
        assert_that(ret['spb'], equal_to(40.0))

    @patch_rest
    def test_sp_busy_idle_util(self):
        paths = ['sp.*.cpu.summary.busyTicks',
                 'sp.*.cpu.summary.idleTicks']
        ret = sp_busy_idle_util(paths, qr_17, qr_34)
        assert_that(ret['spa'], equal_to(22.0))
        assert_that(ret['spb'], equal_to(33.0))

    @patch_rest
    def test_sp_temperature(self):
        path = 'sp.*.platform.storageProcessorTemperature'
        ret = sp_fact(path, qr_17, qr_34)
        assert_that(ret['spa'], equal_to(27))
        assert_that(ret['spb'], equal_to(28))

    def test_only_one_path_set(self):
        assert_that(only_one_path({'abc'}), equal_to('abc'))

    def test_only_one_path_error(self):
        def f():
            return only_one_path(('a', 'b'))

        assert_that(f, raises(ValueError, 'only one path'))

    def test_all_not_none(self):
        assert_that(all_not_none(1, 'a', 2.3), equal_to(True))
        assert_that(all_not_none(1, 'a', None, 2.3), equal_to(False))

    @patch_rest
    def test_total_delta_ps(self):
        path = ['sp.*.physical.disk.*.reads',
                'sp.*.physical.disk.*.writes']
        ret = total_delta_ps(path, qr_6, qr_14)

        delta_reads = (4158667 + 5) - (2966780 + 5)
        delta_writes = (3762339 + 5) - (2679611 + 5)
        diff_time = 163800.0

        expected = (delta_reads + delta_writes) / diff_time
        assert_that(ret['dae_0_1_disk_2'], close_to(expected, 0.01))

        assert_that(total_delta_ps(path, qr_6, qr_14, 'dae_0_1_disk_2'),
                    close_to(expected, 0.01))

    def test_total_delta_ps_error_path(self):
        def f():
            total_delta_ps(['a', 'b', 'c'], None, None)

        assert_that(f, raises(ValueError,
                              'takes in "reads" and "writes" counter.'))

    @patch_rest
    def test_sp_total_delta_ps(self):
        path = ['sp.*.storage.summary.reads',
                'sp.*.storage.summary.writes']
        ret = sp_total_delta_ps(path, qr_17, qr_34)

        diff_time = 180.0
        expected_spa = ((270 - 0) + (306 - 0)) / diff_time
        expected_spb = ((296 - 8) + (324 - 0)) / diff_time
        assert_that(ret['spa'], close_to(expected_spa, 0.01))
        assert_that(ret['spb'], close_to(expected_spb, 0.01))

        assert_that(sp_total_delta_ps(path, qr_17, qr_34, 'spa'),
                    close_to(expected_spa, 0.01))
        assert_that(sp_total_delta_ps(path, qr_17, qr_34, 'spb'),
                    close_to(expected_spb, 0.01))

    def test_sp_total_delta_ps_error_path(self):
        def f():
            sp_total_delta_ps(['a', 'b', 'c'], None, None)

        assert_that(f, raises(ValueError,
                              'takes in "reads" and "writes" counter.'))

    @patch_rest
    def test_system_delta_ps(self):
        path = 'sp.*.storage.summary.reads'
        ret = system_delta_ps(path, qr_17, qr_34)

        diff_time = 180.0
        expected = ((270 - 0) + (296 - 8)) / diff_time
        assert_that(ret['0'], close_to(expected, 0.01))

    def test_system_delta_ps_error_path(self):
        def f():
            system_delta_ps(['a', 'b'], None, None)

        assert_that(f, raises(ValueError, 'takes in one and only one path.'))

    @patch_rest
    def test_system_total_delta_ps(self):
        path = ['sp.*.storage.summary.reads',
                'sp.*.storage.summary.writes']
        ret = system_total_delta_ps(path, qr_17, qr_34)

        delta_total_reads = (270 - 0) + (296 - 8)
        delta_total_writes = (306 - 0) + (324 - 0)
        diff_time = 180.0

        expected = (delta_total_reads + delta_total_writes) / diff_time
        assert_that(ret['0'], close_to(expected, 0.01))

    def test_system_total_delta_ps_error_path(self):
        def f():
            system_total_delta_ps(['a', 'b', 'c'], None, None)

        assert_that(f, raises(ValueError,
                              'takes in "reads" and "writes" counter.'))

    @patch_rest
    def test_disk_response_time(self):
        path = ['sp.*.physical.disk.*.busyTicks',
                'sp.*.physical.disk.*.sumArrivalQueueLength',
                'sp.*.physical.disk.*.reads',
                'sp.*.physical.disk.*.writes',
                'sp.*.physical.coreCount']
        ret = disk_response_time(path, qr_128, qr_130)

        delta_busy_ticks = (151624601961 + 3485236370346) - \
                           (149509372140 + 3436715092500)
        delta_sum_ql = (3532917 + 705) - (3484428 + 705)
        delta_reads = (3532917 + 704) - (3484427 + 704)
        delta_writes = (0 + 0) - (0 + 0)
        delta_size_reads = (3532917 - 3484427) * 10 + (704 - 704) * 10
        delta_size_writes = (0 - 0) * 10 + (0 - 0) * 10

        expected = (delta_busy_ticks * delta_sum_ql) / \
                   ((delta_reads + delta_writes) *
                    (delta_size_reads + delta_size_writes))
        assert_that(ret['dae_0_1_disk_1'], close_to(expected, 0.01))

    def test_disk_response_time_error_path(self):
        def f():
            disk_response_time(['a', 'b', 'c', 'd', 'e', 'f'], None, None)

        assert_that(f, raises(ValueError,
                              'takes in "busyTicks", "sumArrivalQueueLength",'
                              ' "reads", "writes" and "coreCount" counter.'))

    @patch_rest
    def test_disk_queue_length(self):
        path = ['sp.*.physical.disk.*.sumArrivalQueueLength',
                'sp.*.physical.disk.*.reads',
                'sp.*.physical.disk.*.writes']
        ret = disk_queue_length(path, qr_128, qr_130)

        delta_sum_ql = (3532917 + 705) - (3484428 + 705)
        delta_reads = (3532917 + 704) - (3484427 + 704)
        delta_writes = (0 + 0) - (0 + 0)

        expected = delta_sum_ql / (delta_reads + delta_writes)
        assert_that(ret['dae_0_1_disk_1'], close_to(expected, 0.01))

    def test_disk_queue_length_error_path(self):
        def f():
            disk_queue_length(['a', 'b', 'c', 'd'], None, None)

        msg = 'takes in "sumArrivalQueueLength", "reads" and "writes" counter.'
        assert_that(f, raises(ValueError, msg))

    @patch_rest
    def test_lun_response_time(self):
        path = ['sp.*.storage.lun.*.totalIoTime',
                'sp.*.storage.lun.*.reads',
                'sp.*.storage.lun.*.writes']
        ret = lun_response_time(path, qr_128, qr_130)

        delta_sum_ql = (13012626542 + 27073895) - (12782881910 + 26805216)
        delta_reads = (166979 + 0) - (164007 + 0)
        delta_writes = (1618548 + 213) - (1596862 + 213)

        expected = delta_sum_ql / (delta_reads + delta_writes)
        assert_that(ret['sv_4'], close_to(expected, 0.01))

    def test_lun_response_time_error_path(self):
        def f():
            lun_response_time(['a', 'b', 'c', 'd'], None, None)

        msg = 'takes in "totalIoTime", "reads" and "writes" counter.'
        assert_that(f, raises(ValueError, msg))

    @patch_rest
    def test_lun_queue_length(self):
        path = ['sp.*.storage.lun.*.currentIOCount',
                'sp.*.storage.lun.*.busyTime',
                'sp.*.storage.lun.*.idleTime']
        ret = lun_queue_length(path, qr_128, qr_130)

        delta_io = (0 + 0) - (0 + 0)
        delta_busy = (319229585 + 1193127357) - (314779461 + 1176806087)
        delta_idle = (2989316155 + 3185505629) - (4117396013 + 4325740825)

        expected = delta_io * delta_busy / (delta_busy + delta_idle)
        assert_that(ret['sv_4'], close_to(expected, 0.01))

    def test_lun_queue_length_error_path(self):
        def f():
            lun_queue_length(['a', 'b', 'c', 'd'], None, None)

        msg = 'takes in "currentIOCount", "busyTime" and "idleTime" counter.'
        assert_that(f, raises(ValueError, msg))

    @patch_rest
    def test_sp_sum_values(self):
        path = 'sp.*.fastCache.volume.*.readHits'
        ret = sp_sum_values(path, qr_128, qr_130)
        expected_spa = 1000 + 2000 + 3000 + 4000 + 5000
        assert_that(ret['spa'], equal_to(expected_spa))
        expected_spb = 2000 + 4000 + 6000 + 8000 + 10000
        assert_that(ret['spb'], equal_to(expected_spb))

    def test_sp_sum_values_error_path(self):
        def f():
            sp_sum_values(['a', 'b'], None, None)

        assert_that(f, raises(ValueError, 'takes in one and only one path.'))

    @patch_rest
    def test_sp_io_rate(self):
        path = 'sp.*.fastCache.volume.*.readHits'
        ret = sp_io_rate(path, qr_128, qr_130)
        diff_time = 48540.0

        delta_spa = (1000 + 2000 + 3000 + 4000 + 5000) - (
            1000 + 1000 + 1000 + 1000 + 1000)
        expected_spa = delta_spa / diff_time
        assert_that(ret['spa'], close_to(expected_spa, 0.01))

        delta_spb = (2000 + 4000 + 6000 + 8000 + 10000) - (
            2000 + 2000 + 2000 + 2000 + 2000)
        expected_spb = delta_spb / diff_time
        assert_that(ret['spb'], close_to(expected_spb, 0.01))

    def test_sp_io_rate_error_path(self):
        def f():
            sp_io_rate(['a', 'b'], None, None)

        assert_that(f, raises(ValueError, 'takes in one and only one path.'))

    @patch_rest
    def test_byte_rate(self):
        path = ['sp.*.storage.lun.*.readBlocks',
                'sp.*.storage.blockSize']
        ret = byte_rate(path, qr_128, qr_130)

        delta_read_blocks = (10831709 - 10677980)
        block_size = 512
        diff_time = 48540.0

        expected = delta_read_blocks * block_size / diff_time
        assert_that(ret['sv_4'], close_to(expected, 0.01))

    def test_byte_rate_error_path(self):
        def f():
            byte_rate(['a', 'b', 'c'], None, None, None)

        assert_that(f, raises(ValueError,
                              'takes in "Blocks" and "blockSize" counter.'))

    @patch_rest
    def test_total_byte_rate(self):
        path = ['sp.*.physical.disk.*.readBlocks',
                'sp.*.physical.disk.*.writeBlocks',
                'sp.*.physical.blockSize']
        ret = total_byte_rate(path, qr_128, qr_130)

        delta_read_byte = (7235384336 - 7136076816) * 512 + \
                          (1348912 - 1348912) * 512
        delta_write_byte = (1000 - 10) * 512 + (1100 - 20) * 512
        diff_time = 48540.0

        expected = (delta_read_byte + delta_write_byte) / diff_time
        assert_that(ret['dae_0_1_disk_1'], close_to(expected, 0.01))

    def test_total_byte_rate_error_path(self):
        def f():
            total_byte_rate(['a', 'b', 'c', 'd'], None, None)

        msg = 'takes in "readBlocks", "writeBlocks" and "blockSize" counter.'
        assert_that(f, raises(ValueError, msg))

    @patch_rest
    def test_sp_byte_rate(self):
        path = ['sp.*.storage.summary.readBlocks',
                'sp.*.storage.blockSize']
        ret = sp_byte_rate(path, qr_128, qr_130)

        delta_blocks = 11450335 - 11289244
        block_size = 512
        diff_time = 48540.0

        expected = delta_blocks * block_size / diff_time
        assert_that(ret['spa'], close_to(expected, 0.01))

    def test_sp_byte_rate_error_path(self):
        def f():
            sp_byte_rate(['a', 'b', 'c'], None, None)

        assert_that(f, raises(ValueError,
                              'takes in "Blocks" and "blockSize" counter.'))

    @patch_rest
    def test_sp_total_byte_rate(self):
        path = ['sp.*.storage.summary.readBlocks',
                'sp.*.storage.summary.writeBlocks',
                'sp.*.storage.blockSize']
        ret = sp_total_byte_rate(path, qr_128, qr_130)

        delta_blocks_1 = 11450335 - 11289244
        delta_blocks_2 = 33497820 - 32954971
        block_size = 512
        diff_time = 48540.0

        expected = (delta_blocks_1 + delta_blocks_2) * block_size / diff_time
        assert_that(ret['spa'], close_to(expected, 0.01))

    def test_sp_total_byte_rate_error_path(self):
        def f():
            sp_total_byte_rate(['a', 'b', 'c', 'd'], None, None)

        msg = 'takes in "readBlocks", "writeBlocks" and "blockSize" counter.'
        assert_that(f, raises(ValueError, msg))

    @patch_rest
    def test_system_byte_rate(self):
        path = ['sp.*.storage.summary.readBlocks',
                'sp.*.storage.blockSize']
        ret = system_byte_rate(path, qr_128, qr_130)

        delta_bytes = (11450335 - 11289244) * 512 + (183144 - 183137) * 512
        diff_time = 48540.0

        expected = delta_bytes / diff_time
        assert_that(ret['0'], close_to(expected, 0.01))

    def test_system_byte_rate_error_path(self):
        def f():
            system_byte_rate(['a', 'b', 'c'], None, None)

        assert_that(f, raises(ValueError,
                              'takes in "Blocks" and "blockSize" counter.'))

    @patch_rest
    def test_system_total_byte_rate(self):
        path = ['sp.*.storage.summary.readBlocks',
                'sp.*.storage.summary.writeBlocks',
                'sp.*.storage.blockSize']
        ret = system_total_byte_rate(path, qr_128, qr_130)

        delta_byte_1 = ((11450335 - 11289244) + (183144 - 183137)) * 512
        delta_byte_2 = ((33497820 - 32954971) + (70507 - 70507)) * 512
        diff_time = 48540.0

        expected = (delta_byte_1 + delta_byte_2) / diff_time
        assert_that(ret['0'], close_to(expected, 0.01))

    def test_system_total_byte_rate_error_path(self):
        def f():
            system_total_byte_rate(['a', 'b', 'c', 'd'], None, None)

        msg = 'takes in "readBlocks", "writeBlocks" and "blockSize" counter.'
        assert_that(f, raises(ValueError, msg))


class UnityMetricConfigTest(TestCase):
    def test_default_calculator(self):
        mc = UnityMetricConfig({'name': 'test'})
        assert_that(mc.calculator, equal_to(calculator.delta_ps))

    def test_path_str(self):
        mc = UnityMetricConfig({'name': 'test', 'paths': 'abc'})
        assert_that(len(mc.paths), equal_to(1))
        assert_that(mc.paths, has_item('abc'))

    def test_path_collection(self):
        mc = UnityMetricConfig({'name': 'test', 'paths': ['a', 'b']})
        assert_that(mc.paths, has_items('a', 'b'))


class UnityMetricConfigParserTest(TestCase):
    config = UnityMetricConfigParser()

    def test_get_configs(self):
        disk_config = self.config.get_config('UnityDisk')
        assert_that(disk_config.metric_names(),
                    has_items('read_mbps', 'write_iops'))



# coding=utf-8
# Copyright (c) 2017 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import threading
import time
from collections import OrderedDict

from storops.lib.common import singleton
from storops.unity.resource.job import UnityJobList, wait_job_completion

log = logging.getLogger(__name__)


@singleton
class JobHelper(object):
    def __init__(self, cli, interval=3):
        self._cli = cli
        self._interval = interval
        self.jobs = OrderedDict()
        self.started = False

    def add_job(self, job):
        if job.id not in self.jobs:
            self.jobs[job.id] = job

    def remove_job(self, job):
        if job.id in self.jobs:
            del self.jobs[job.id]

    def get_job(self, job):
        if job.id in self.jobs:
            return self.jobs[job.id]

    def wait_job(self, job, timeout=3600, interval=3):
        def update_func(job):
            return self.get_job(job)

        try:
            self.add_job(job)
            job = wait_job_completion(job, timeout=timeout, interval=interval,
                                      update_func=update_func)
            return job
        finally:
            self.remove_job(job)

    def _query_jobs(self):
        if len(self.jobs) > 0:
            filters = {'id': list(self.jobs.keys())}
            log.debug('Query jobs from Unity, filter: {}'.format(filters))
            jobs = UnityJobList.get(cli=self._cli, **filters)
            log.debug('Query result: {}.'.format(jobs))
            for job in jobs:
                if hasattr(job, 'id'):
                    self.jobs[job.id] = job

    def query_jobs(self):
        try:
            while True:
                self._query_jobs()
                time.sleep(self._interval)
        except Exception as err:
            log.exception(
                'Job helper (thread: {}) exited with exception: {}'.format(
                    self.t.name, str(err)))
            self.started = False
            raise err

    def start(self):
        if self.started:
            return

        self.t = threading.Thread(name='job-helper-daemon',
                                  target=self.query_jobs)
        self.t.setDaemon(True)
        self.t.start()
        log.info('Job helper (thread: {}) started.'.format(self.t.name))
        self.started = True


def get_job_helper(cli):
    jh = JobHelper(cli)
    if not jh.started:
        jh.start()
    return jh



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from storops import exception as ex
from storops.unity.resource.job import UnityJob


class MockJobHelper(object):
    def __init__(self, cli, interval=5):
        self._cli = cli
        self.started = True

    def wait_job(self, job, async_timeout, async_interval):
        if job.id == 'N-3078':
            return UnityJob(_id=job.id, cli=self._cli)
        if job.id == 'N-3079':
            ret_job = UnityJob(_id=job.id, cli=self._cli)
            raise ex.JobStateError(ret_job)
        if job.id == 'N-3080':
            raise ex.JobTimeoutException()



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

import storops.exception
from storops.lib import job_helper
from storops.lib.common import instance_cache
from storops.lib.resource import ResourceList
from storops.lib.version import version
from storops.unity.calculator import calculators
from storops.unity.client import UnityClient
from storops.unity.enums import UnityEnum, DNSServerOriginEnum
from storops.unity.resource import UnityResource, UnityResourceList, \
    UnitySingletonResource, UnityAttributeResource
from storops.unity.resource.alert import UnityAlertList
from storops.unity.resource.alert_config_snmp import \
    UnityAlertConfigSNMPTarget, UnityAlertConfigSNMPTargetList
from storops.unity.resource.cifs_server import UnityCifsServerList
from storops.unity.resource.cifs_share import UnityCifsShareList
from storops.unity.resource.disk import UnityDiskList, UnityDiskGroupList
from storops.unity.resource.dns_server import UnityFileDnsServerList
from storops.unity.resource.filesystem import UnityFileSystemList
from storops.unity.resource.host import UnityHost, UnityHostList, \
    UnityHostIpPortList, UnityHostInitiatorList
from storops.unity.resource.host_container import UnityHostContainer, \
    UnityHostContainerList
from storops.unity.resource.import_session import UnityImportSessionList
from storops.unity.resource.interface import UnityFileInterfaceList
from storops.unity.resource.job import UnityJobList
from storops.unity.resource.lun import UnityLunList
from storops.unity.resource.metric import UnityMetricRealTimeQuery
from storops.unity.resource.move_session import UnityMoveSessionList
from storops.unity.resource.nas_server import UnityNasServerList
from storops.unity.resource.nfs_server import UnityNfsServerList
from storops.unity.resource.nfs_share import UnityNfsShareList
from storops.unity.resource.pool import UnityPoolList, UnityPool
from storops.unity.resource.port import UnityEthernetPortList, \
    UnityIscsiPortalList, UnityFcPortList, UnityIoModuleList
from storops.unity.resource.port import UnityIpPortList, UnityIoLimitPolicy, \
    UnityIoLimitPolicyList, UnityLinkAggregationList, UnityIscsiPortal, \
    UnitySasPortList, UnityIscsiNodeList
from storops.unity.resource.remote_interface import UnityRemoteInterfaceList
from storops.unity.resource.remote_system import UnityRemoteSystemList, \
    UnityRemoteSystem
from storops.unity.resource.replication_interface import \
    UnityReplicationInterfaceList, UnityReplicationInterface
from storops.unity.resource.replication_session import \
    UnityReplicationSessionList
from storops.unity.resource.snap import UnitySnapList
from storops.unity.resource.snap_schedule import UnitySnapScheduleList, \
    UnitySnapSchedule
from storops.unity.resource.sp import UnityStorageProcessorList
from storops.unity.resource.storage_resource import UnityStorageResourceList
from storops.unity.resource.cg import UnityConsistencyGroup, \
    UnityConsistencyGroupList
from storops.unity.resource.tenant import UnityTenant, UnityTenantList
from storops.unity.resource.vmware import UnityCapabilityProfile, \
    UnityCapabilityProfileList, UnityDataStore

__author__ = 'Jay Xu, Cedric Zhuang'

LOG = logging.getLogger(__name__)


class UnitySystem(UnitySingletonResource):
    def __init__(self, host=None, username=None, password=None,
                 port=443, cli=None, verify=False, retries=None,
                 cache_interval=0):
        super(UnitySystem, self).__init__(cli=cli)
        if cli is None:
            self._cli = UnityClient(host, username, password, port,
                                    verify=verify, retries=retries,
                                    cache_interval=cache_interval)
        else:
            self._cli = cli

        # Start job helper to handle async operations.
        job_helper.get_job_helper(self._cli)

    @classmethod
    def get_resource_class(cls):
        return cls

    def get_capability_profile(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityCapabilityProfileList,
                                   _id=_id, name=name, **filters)

    def get_sp(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityStorageProcessorList, _id=_id,
                                   name=name, **filters)

    def get_iscsi_portal(self, _id=None, **filters):
        return self._get_unity_rsc(UnityIscsiPortalList, _id=_id, **filters)

    def get_iscsi_node(self, _id=None, **filters):
        return self._get_unity_rsc(UnityIscsiNodeList, _id=_id, **filters)

    def get_fc_port(self, _id=None, **filters):
        return self._get_unity_rsc(UnityFcPortList, _id=_id, **filters)

    def get_ethernet_port(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityEthernetPortList, _id=_id,
                                   name=name, **filters)

    def create_host(self, name, host_type=None, desc=None, os=None,
                    tenant=None):
        host = UnityHostList.get(self._cli, name=name)
        if host:
            raise storops.exception.UnityHostNameInUseError()
        else:
            host = UnityHost.create(self._cli, name, host_type=host_type,
                                    desc=desc, os=os, tenant=tenant)

        return host

    def get_initiator(self, _id=None, **filters):
        return self._get_unity_rsc(UnityHostInitiatorList, _id=_id, **filters)

    def get_lun(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityLunList, _id=_id, name=name, **filters)

    def create_pool(self, name, raid_groups, description=None, **kwargs):
        """Create pool based on RaidGroupParameter.

        :param name: pool name
        :param raid_groups: a list of *RaidGroupParameter*
        :param description: pool description
        :param alert_threshold: Threshold at which the system will generate
               alerts about the free space in the pool, specified as
               a percentage.
        :param is_harvest_enabled:
               True - Enable pool harvesting for the pool.
               False - Disable pool harvesting for the pool.
        :param is_snap_harvest_enabled:
               True - Enable snapshot harvesting for the pool.
               False - Disable snapshot harvesting for the pool.
        :param pool_harvest_high_threshold: Pool used space high threshold at
               which the system will automatically starts to delete snapshots
               in the pool
        :param pool_harvest_low_threshold: Pool used space low threshold under
               which the system will automatically stop deletion of snapshots
               in the pool
        :param snap_harvest_high_threshold: Snapshot used space high threshold
               at which the system automatically starts to delete snapshots
               in the pool
        :param snap_harvest_low_threshold: Snapshot used space low threshold
               below which the system will stop automatically deleting
               snapshots in the pool
        :param is_fast_cache_enabled:
               True - FAST Cache will be enabled for this pool.
               False - FAST Cache will be disabled for this pool.
        :param is_fastvp_enabled:
               True - Enable scheduled data relocations for the pool.
               False - Disable scheduled data relocations for the pool.
        :param pool_type:
               StoragePoolTypeEnum.TRADITIONAL - Create traditional pool.
               StoragePoolTypeEnum.DYNAMIC - Create dynamic pool. (default)
        """
        return UnityPool.create(self._cli, name=name, description=description,
                                raid_groups=raid_groups, **kwargs)

    def get_pool(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityPoolList, _id=_id, name=name,
                                   **filters)

    def get_nas_server(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityNasServerList, _id=_id, name=name,
                                   **filters)

    def get_cifs_server(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityCifsServerList, _id=_id, name=name,
                                   **filters)

    def get_nfs_server(self, _id=None, **filters):
        return self._get_unity_rsc(UnityNfsServerList, _id=_id, **filters)

    def get_snap(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnitySnapList, _id=_id, name=name,
                                   **filters)

    def create_nas_server(self, name, sp=None, pool=None, is_repl_dst=None,
                          multi_proto=None, tenant=None):
        if sp is None:
            sp = self._auto_balance_sp()

        return sp.create_nas_server(name, pool,
                                    is_repl_dst=is_repl_dst,
                                    multi_proto=multi_proto, tenant=tenant)

    def _auto_balance_sp(self):
        sp_list = self.get_sp()
        if len(sp_list) < 2:
            LOG.debug('spb not present. pick spa to host nas server.')
            return sp_list.first_item

        servers = self.get_nas_server()
        spa, spb = sp_list
        servers_on_spa = servers.shadow_copy(home_sp=spa)
        sp = spb if len(servers_on_spa) * 2 > len(servers) else spa
        LOG.debug('pick %s to balance of spa and spb to host nas servers.',
                  sp.get_id())
        return sp

    def get_ip_port(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityIpPortList, _id=_id, name=name,
                                   **filters)

    @version(">=4.1")
    def get_file_port(self):
        """Returns ports list can be used by File

        File ports includes ethernet ports and link aggregation ports.
        """
        eths = self.get_ethernet_port(bond=False)
        las = self.get_link_aggregation()
        return eths + las

    @version("<4.1")  # noqa
    def get_file_port(self):
        """Returns ports list can be used by File

        File ports includes ethernet ports and link aggregation ports.
        """
        eths = self.get_ethernet_port()
        return eths.list

    def get_file_interface(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityFileInterfaceList, _id=_id, name=name,
                                   **filters)

    def create_iscsi_portal(self, ethernet_port, ip, netmask=None,
                            v6_prefix_len=None, vlan=None, gateway=None):
        return UnityIscsiPortal.create(
            cli=self._cli, ethernet_port=ethernet_port, ip=ip, netmask=netmask,
            v6_prefix_len=v6_prefix_len, vlan=vlan, gateway=gateway)

    def get_dns_server(self, _id=None, **filters):
        return self._get_unity_rsc(UnityFileDnsServerList, _id=_id, **filters)

    def get_filesystem(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityFileSystemList, _id=_id, name=name,
                                   **filters)

    def get_cifs_share(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityCifsShareList, _id=_id, name=name,
                                   allow_dup=True, **filters)

    def get_nfs_share(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityNfsShareList, _id=_id, name=name,
                                   allow_dup=True, **filters)

    def get_host(self, _id=None, name=None, address=None, **filters):
        ret = UnityHostList.get(self._cli, name='not found')
        if address:
            host_ip_ports = UnityHostIpPortList.get(self._cli, address=address)
            if host_ip_ports:
                ret = host_ip_ports[0].host
        else:
            ret = self._get_unity_rsc(UnityHostList, _id=_id, name=name,
                                      **filters)
        return ret

    def get_io_limit_policy(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityIoLimitPolicyList, _id=_id, name=name,
                                   **filters)

    def create_io_limit_policy(self, name, max_iops=None, max_kbps=None,
                               policy_type=None, is_shared=None,
                               description=None, max_iops_density=None,
                               max_kbps_density=None, burst_rate=None,
                               burst_time=None, burst_frequency=None):
        return UnityIoLimitPolicy.create(
            self._cli, name, max_iops=max_iops, max_kbps=max_kbps,
            policy_type=policy_type, is_shared=is_shared,
            description=description,
            max_iops_density=max_iops_density,
            max_kbps_density=max_kbps_density,
            burst_rate=burst_rate,
            burst_time=burst_time,
            burst_frequency=burst_frequency)

    def get_cg(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityConsistencyGroupList, _id=_id,
                                   name=name, **filters)

    def create_cg(self, name, description=None, lun_add=None, hosts=None):
        return UnityConsistencyGroup.create(
            self._cli, name, description=description, lun_add=lun_add,
            hosts=hosts)

    def get_doc(self, resource):
        if isinstance(resource, (UnityResource, UnityEnum)):
            clz = resource.__class__
        else:
            clz = resource
        return self._cli.get_doc(clz)

    @version('>=4.1')
    def create_tenant(self, name, uuid=None, vlans=None):
        return UnityTenant.create(self._cli, name, uuid=uuid, vlans=vlans)

    @version('>=4.1')
    def get_tenant(self, _id=None, **filters):
        return self._get_unity_rsc(UnityTenantList, _id=_id, **filters)

    @version('>=4.1')
    def get_tenant_use_vlan(self, vlan):
        tenant = self.get_tenant(vlans=[vlan])
        if len(tenant) == 0:
            return None
        else:
            return tenant[0]

    @property
    @instance_cache
    def _system_time(self):
        return UnitySystemTime(self._cli)

    @property
    def system_time(self):
        return self._system_time.time

    def set_system_time(self, new_time=None):
        return self._system_time.set(new_time)

    @property
    @instance_cache
    def _ntp_server(self):
        return UnityNtpServer(self._cli)

    @property
    def ntp_server(self):
        return self._ntp_server.addresses

    def add_ntp_server(self, *addresses):
        return self._ntp_server.add(*addresses)

    def remove_ntp_server(self, *addresses):
        return self._ntp_server.remove(*addresses)

    def clear_ntp_server(self):
        return self._ntp_server.clear()

    @property
    @instance_cache
    def dns_server(self):
        return UnityDnsServer(self._cli)

    def add_dns_server(self, *addresses):
        return self.dns_server.add(*addresses)

    def remove_dns_server(self, *addresses):
        return self.dns_server.remove(*addresses)

    def clear_dns_server(self, use_dhcp=None):
        """ Clear the DNS server settings.

        :param use_dhcp: default to True, clear all settings and fallback to
                         use DHCP settings.
        :return: last settings
        """
        return self.dns_server.clear(use_dhcp)

    def get_disk(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityDiskList, _id=_id, name=name,
                                   **filters)

    def get_disk_group(self, _id=None):
        return self._get_unity_rsc(UnityDiskGroupList, _id=_id)

    def get_dae(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityDaeList, _id=_id, name=name, **filters)

    def get_dpe(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityDpeList, _id=_id, name=name, **filters)

    def get_sas_port(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnitySasPortList, _id=_id, name=name, **filters)

    def get_io_module(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityIoModuleList, _id=_id, name=name, **filters)

    def get_lcc(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityLccList, _id=_id, name=name, **filters)

    def get_memory_module(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityMemoryModuleList, _id=_id, name=name, **filters)

    def get_fan(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityFanList, _id=_id, name=name, **filters)

    def get_power_supply(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityPowerSupplyList, _id=_id, name=name, **filters)

    def get_battery(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityBatteryList, _id=_id, name=name, **filters)

    def get_ssc(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnitySscList, _id=_id, name=name, **filters)

    def get_ssd(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnitySsdList, _id=_id, name=name, **filters)

    def get_license(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityLicenseList, _id=_id, name=name, **filters)

    def get_feature(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityFeatureList, _id=_id, name=name, **filters)

    def get_mgmt_interface(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityMgmtInterfaceList, _id=_id, **filters)

    def get_system_capacity(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnitySystemCapacityList, _id=_id, **filters)

    def get_move_session(self, _id=None, **filters):
        return self._get_unity_rsc(UnityMoveSessionList, _id=_id, **filters)

    @property
    @instance_cache
    def info(self):
        return UnityBasicSystemInfo.get(cli=self._cli)

    @version('>=4.1')
    def get_link_aggregation(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityLinkAggregationList, _id=_id,
                                   name=name,
                                   **filters)

    def enable_perf_stats(self, interval=None, rsc_clz_list=None):
        if interval is None:
            interval = 60
        if rsc_clz_list is None:
            rsc_list_collection = self._default_rsc_list_with_perf_stats()
            rsc_clz_list = ResourceList.get_rsc_clz_list(rsc_list_collection)

        def get_real_time_query_list():
            paths = calculators.get_all_paths(rsc_clz_list)
            return UnityMetricRealTimeQuery.get_query_list(
                self._cli, interval, paths=paths)

        def f():
            query_list = get_real_time_query_list()
            if query_list:
                paths = calculators.get_all_paths(rsc_clz_list)
                ret = query_list.get_query_result(paths)
            else:
                ret = None
            return ret

        queries = get_real_time_query_list()
        self._cli.enable_perf_metric(interval, f, rsc_clz_list)
        return queries

    def disable_perf_stats(self):
        self._cli.disable_perf_metric()

    def is_perf_stats_enabled(self):
        return self._cli.is_perf_metric_enabled()

    def get_metric_query_result(self, query_id):
        return UnityMetricRealTimeQuery(
            cli=self._cli, _id=query_id).get_query_result()

    def add_metric_record(self, record):
        self._cli.add_metric_record(record)

    def enable_persist_perf_stats(self):
        rsc_list = self._default_rsc_list_with_perf_stats()
        self._cli.persist_perf_stats(rsc_list)

    def disable_persist_perf_stats(self):
        self._cli.persist_perf_stats(None)

    def _default_rsc_list_with_perf_stats(self):
        return (self.get_sp(),
                self.get_lun(),
                self.get_filesystem(),
                self.get_disk(inserted=True),
                self.get_fc_port(),
                self.get_iscsi_node(),
                self.update(),
                )

    def is_perf_stats_persisted(self):
        return self._cli.is_perf_stats_persisted()

    def upload_license(self, license):
        files = {'filename': license}
        resp = self._cli.rest_post('/upload/license', files=files)
        resp.raise_if_err()

    def get_remote_system(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityRemoteSystemList, _id=_id, name=name, **filters)

    def create_remote_system(self, management_address,
                             local_username=None, local_password=None,
                             remote_username=None, remote_password=None,
                             connection_type=None):
        """
        Configures a remote system for remote replication.

        :param management_address: the management IP address of the remote
            system.
        :param local_username: administrative username of local system.
        :param local_password: administrative password of local system.
        :param remote_username: administrative username of remote system.
        :param remote_password: administrative password of remote system.
        :param connection_type: `ReplicationCapabilityEnum`. Replication
            connection type to the remote system.
        :return: the newly created remote system.
        """
        return UnityRemoteSystem.create(self._cli, management_address,
                                        local_username=local_username,
                                        local_password=local_password,
                                        remote_username=remote_username,
                                        remote_password=remote_password,
                                        connection_type=connection_type)

    def get_remote_interface(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityRemoteInterfaceList, _id=_id, name=name, **filters)

    def get_replication_interface(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityReplicationInterfaceList, _id=_id, name=name, **filters)

    def create_replication_interface(self, sp, ip_port, ip_address,
                                     netmask=None, v6_prefix_length=None,
                                     gateway=None, vlan_id=None):
        """
        Creates a replication interface.

        :param sp: `UnityStorageProcessor` object. Storage processor on which
            the replication interface is running.
        :param ip_port: `UnityIpPort` object. Physical port or link aggregation
            on the storage processor on which the interface is running.
        :param ip_address: IP address of the replication interface.
        :param netmask: IPv4 netmask for the replication interface, if it uses
            an IPv4 address.
        :param v6_prefix_length: IPv6 prefix length for the interface, if it
            uses an IPv6 address.
        :param gateway: IPv4 or IPv6 gateway address for the replication
            interface.
        :param vlan_id: VLAN identifier for the interface.
        :return: the newly create replication interface.
        """
        return UnityReplicationInterface.create(
            self._cli, sp, ip_port, ip_address, netmask=netmask,
            v6_prefix_length=v6_prefix_length, gateway=gateway,
            vlan_id=vlan_id)

    def get_replication_session(self, _id=None, name=None,
                                src_resource_id=None, dst_resource_id=None,
                                **filters):
        return self._get_unity_rsc(
            UnityReplicationSessionList, _id=_id, name=name,
            src_resource_id=src_resource_id, dst_resource_id=dst_resource_id,
            **filters)

    def get_import_session(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(
            UnityImportSessionList, _id=_id, name=name, **filters)

    def create_alert_snmp_config(self, target_address, auth_protocol=None,
                                 username=None, auth_password=None,
                                 priv_protocol=None, priv_password=None,
                                 snmp_version=None, community=None):
        if not (community or username):
            raise storops.exception.UnityPolicyInvalidParametersError()
        alert = UnityAlertConfigSNMPTarget.create(self._cli, target_address,
                                                  auth_protocol=auth_protocol,
                                                  username=username,
                                                  auth_password=auth_password,
                                                  priv_protocol=priv_protocol,
                                                  priv_password=priv_password,
                                                  snmp_version=snmp_version,
                                                  community=community)
        return alert

    def get_alert_snmp_config(self, _id=None, **filters):
        return self._get_unity_rsc(UnityAlertConfigSNMPTargetList,
                                   _id=_id, **filters)

    def get_job(self, _id=None, **filters):
        return self._get_unity_rsc(UnityJobList, _id=_id, **filters)

    def get_alert(self, _id=None, **filters):
        return self._get_unity_rsc(UnityAlertList, _id=_id, **filters)

    def create_snap_schedule(self, name, rules,
                             is_sync_replicated=None,
                             skip_sync_to_remote_system=None):
        """ Creates snapshot schedule.
        :param name: Name of new schedule.
        :param rules: Rules that apply to the snapshot schedule, as defined by
            the `UnitySnapScheduleRule` resource type. Each value is a
            `UnitySnapScheduleRule` instance.
        :param is_sync_replicated: Indicates that all operations on the
            snapshot schedule will be synchronously replicated to the peer
            system.
        :param skip_sync_to_remote_system: For internal system use only. Skip
            the invocation of the snapshot schedule creation method with the
            same parameters on the remote system. If the method is invoked by
            the user, this attribute is not required. The value should be true
            when the operation is executed from a remote system, to prevent
            from syncing it back to the original storage.
        :return: the created `UnitySnapSchedule` instance.
        """
        schedule = UnitySnapScheduleList.get(self._cli, name=name)
        if schedule:
            raise storops.exception.UnitySnapScheduleNameInUseError()
        else:
            schedule = UnitySnapSchedule.create(
                self._cli, name, rules, is_sync_replicated=is_sync_replicated,
                skip_sync_to_remote_system=skip_sync_to_remote_system)

        return schedule

    def get_snap_schedule(self, _id=None, **filters):
        return self._get_unity_rsc(UnitySnapScheduleList, _id=_id, **filters)

    def create_capability_profile(self, name, pool, description=None,
                                  usage_tags=None, constraints=None):
        return UnityCapabilityProfile.create(self._cli, name=name,
                                             pool=pool,
                                             description=description,
                                             usage_tags=usage_tags,
                                             constraints=constraints)

    def get_vvol_datastore(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityStorageResourceList,
                                   _id=_id, name=name, **filters)

    def create_vvol_datastore(
            self, name, vvol_datastore_type, description=None, hosts=None,
            capability_profiles=None):
        return UnityDataStore.create(
            self._cli, name, vvol_datastore_type=vvol_datastore_type,
            description=description, hosts=hosts,
            capability_profiles=capability_profiles)

    def get_host_container(self, _id=None, name=None, **filters):
        return self._get_unity_rsc(UnityHostContainerList,
                                   _id=_id, name=name, **filters)

    def recommend_host_container(self, address, username, password):
        return UnityHostContainer.recommend(self._cli, address, username,
                                            password)

    def add_host_container(self, service_type, target_name, target_address,
                           username, password, description=None,
                           local_username=None, local_password=None,
                           potential_hosts=None):
        return UnityHostContainer.create(
            self._cli, service_type, target_name, target_address, username,
            password,
            description=description,
            local_username=local_username, local_password=local_password,
            potential_hosts=potential_hosts
        )


class UnitySystemList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnitySystem


class UnityDpe(UnityResource):
    pass


class UnityDpeList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityDpe


class UnityVirusChecker(UnityResource):
    pass


class UnityVirusCheckerList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityVirusChecker


class UnityBasicSystemInfo(UnitySingletonResource):
    pass


class UnityBasicSystemInfoList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityBasicSystemInfo


class UnitySystemTime(UnitySingletonResource):
    def set(self, new_time, reboot=None):
        old_time = self.time

        if reboot:
            reboot = 1
        else:
            reboot = 0

        time_str = new_time.strftime('%Y-%m-%dT%H:%M:%S.000Z')
        resp = self.modify(time=time_str, rebootPrivilege=reboot)
        resp.raise_if_err()
        self.update()
        return old_time


class _UnityNtpDnsServerCommonBase(UnitySingletonResource):
    def _do_modify(self, *addresses):
        raise NotImplementedError('Should be implemented by child classes.')

    def set(self, *addresses):
        exists = self.addresses
        resp = self._do_modify(*addresses)
        resp.raise_if_err()
        self.update()
        return exists

    def add(self, *addresses):
        new_addresses = set(self.addresses).union(set(addresses))
        return self.set(*new_addresses)

    def remove(self, *addresses):
        new_addresses = set(self.addresses) - set(addresses)
        return self.set(*new_addresses)

    def clear(self):
        return self.set()


class UnityNtpServer(_UnityNtpDnsServerCommonBase):
    def _do_modify(self, *addresses):
        return self.modify(addresses=sorted(addresses), rebootPrivilege=0)


class UnityDnsServer(_UnityNtpDnsServerCommonBase):
    def _do_modify(self, *addresses):
        return self.modify(addresses=sorted(addresses))

    def clear(self, use_dhcp=None):
        if use_dhcp is None:
            use_dhcp = True

        if use_dhcp:
            origin = DNSServerOriginEnum.DHCP
        else:
            origin = None

        exists = self.addresses
        resp = self.modify(addresses=[], origin=origin)
        resp.raise_if_err()
        self.update()
        return exists


class UnityBattery(UnityResource):
    @classmethod
    def get_nested_properties(cls):
        return (
            'parent_storage_processor.name',
        )


class UnityBatteryList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityBattery


class UnityDae(UnityResource):
    pass


class UnityDaeList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityDae


class UnityLcc(UnityResource):
    pass


class UnityLccList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityLcc


class UnityMemoryModule(UnityResource):
    pass


class UnityMemoryModuleList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityMemoryModule


class UnityPowerSupply(UnityResource):
    @classmethod
    def get_nested_properties(cls):
        return (
            'parent_dae.name',
            'parent_dpe.name',
            'storage_processor.name',
        )


class UnityPowerSupplyList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityPowerSupply


class UnitySsc(UnityResource):
    pass


class UnitySscList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnitySsc


class UnitySsd(UnityResource):
    pass


class UnitySsdList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnitySsd


class UnityFan(UnityResource):
    @classmethod
    def get_nested_properties(cls):
        return (
            'parent_dae.name',
            'parent_dpe.name',
        )


class UnityFanList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityFan


class UnityLicense(UnityResource):
    pass


class UnityLicenseList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityLicense


class UnityFeature(UnityResource):
    pass


class UnityFeatureList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityFeature


class UnityMgmtInterface(UnityResource):
    pass


class UnityMgmtInterfaceList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityMgmtInterface


class UnitySystemCapacity(UnityResource):
    pass


class UnitySystemCapacityList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnitySystemCapacity


class UnitySystemTierCapacity(UnityAttributeResource):
    pass


class UnitySystemTierCapacityList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnitySystemTierCapacity



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import re

from storops.exception import check_nas_cmd_error
from storops.vnx.enums import VNXShareType
from storops.vnx.resource.mover import VNXMover
from storops.vnx.resource import VNXResource, VNXCliResourceList

__author__ = 'Cedric Zhuang'

LOG = logging.getLogger(__name__)


class VNXVdmInterface(object):
    def __init__(self, name, share_type):
        self.name = name
        if 'vdm' in share_type:
            share_type = 'nfs'
        self.share_type = VNXShareType.parse(share_type)


class VNXVdmList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXVdm

    def _get_raw_resource(self):
        return self._cli.get_vdm()


class VNXVdm(VNXResource):
    def __init__(self, name=None, vdm_id=None, cli=None):
        super(VNXVdm, self).__init__()
        self._name = name
        self._vdm_id = vdm_id
        self._cli = cli

    def get_vdm_id(self):
        if self._vdm_id is not None:
            ret = self._vdm_id
        else:
            ret = self.vdm_id
        return ret

    def get_mover_id(self):
        return self.get_vdm_id()

    @staticmethod
    def get(cli, name=None, vdm_id=None):
        if name is not None or vdm_id is not None:
            ret = VNXVdm(name=name, vdm_id=vdm_id, cli=cli)
        else:
            ret = VNXVdmList(cli)
        return ret

    @property
    def is_vdm(self):
        return True

    def _get_raw_resource(self):
        if self._vdm_id is not None:
            resp = self._cli.get_vdm(self._vdm_id)
        else:
            resp = self._cli.get_vdm()
            resp.filter_object(name=self._name)
        return resp

    @classmethod
    def create(cls, cli, mover_id, name, pool_id=None):
        resp = cli.create_vdm(mover_id, name, pool_id)
        resp.raise_if_err()
        return VNXVdm(name=name, cli=cli)

    def delete(self):
        resp = self._cli.delete_vdm(self.get_vdm_id())
        resp.raise_if_err()
        return resp

    def get_interfaces(self):
        ret = []

        re_pattern = (r'Interfaces to services mapping:'
                      r'\s*(?P<interfaces>(\s*interface=.*)*)')

        out = self._cli.get_dm_interfaces(name=self._get_name(),
                                          is_vdm=True)

        m = re.search(re_pattern, out)
        if m:
            if_list = m.group('interfaces').split('\n')
            for i in if_list:
                m_if = re.search(r'\s*interface=(?P<if>.*)\s*:'
                                 r'\s*(?P<type>.*)\s*', i)
                if m_if:
                    if_name = m_if.group('if').strip()
                    share_type = m_if.group('type')
                    ret.append(VNXVdmInterface(if_name, share_type))
        return ret

    @property
    def mover(self):
        return VNXMover(mover_id=self.mover_id, cli=self._cli)

    def create_interface(self, device, ip, net_mask, vlan_id=0, name=None):
        return self.mover.create_interface(device, ip, net_mask, vlan_id, name)

    def delete_interface(self, ip):
        return self.mover.delete_interface(ip)

    @property
    def physical_devices(self):
        return self.mover.physical_device

    @property
    def fc_devices(self):
        return self.mover.fc_devices

    @property
    def ethernet_devices(self):
        return self.mover.ethernet_devices

    def attach_nfs_interface(self, if_name):
        out = self._cli.attach_nfs_interface(
            if_name, vdm_name=self._get_name())
        check_nas_cmd_error(out)
        return out

    def detach_nfs_interface(self, if_name):
        out = self._cli.detach_nfs_interface(
            if_name, vdm_name=self._get_name())
        check_nas_cmd_error(out)
        return out



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import os
from datetime import datetime

from storops.exception import VNXPerMonNotEnabledError
from storops.lib.common import instance_cache, clear_instance_cache, \
    get_local_folder
from storops.lib.metric import MetricsDumper
from storops.lib.resource import Resource, ResourceList
from storops.vnx.calculator import calculators
from storops.vnx.parsers import get_vnx_parser

__author__ = 'Cedric Zhuang'


class VNXResource(Resource):
    @classmethod
    def _get_parser(cls):
        return get_vnx_parser(cls.__name__)

    @classmethod
    def resource_class_name(cls):
        return cls._get_parser().resource_class_name

    def _get_value_by_key(self, item):
        ret = super(VNXResource, self)._get_value_by_key(item)
        if ret is None and self._get_parser() is not None:
            prop = self._get_parser().get_property(item)
            if prop and prop.is_resource_list_clazz():
                ret = tuple()
        return ret

    @property
    def system_version(self):
        if self._cli is None:
            return None
        return self._cli.system_version


class _WithPoll(object):
    def __init__(self, r):
        self._resource = r
        self._orig_poll = self._resource.poll

    def __enter__(self):
        pass

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type, exc_val, exc_tb):
        # return None, do not handle inner exception
        self._resource.poll = self._orig_poll


class VNXCliResource(VNXResource):
    def __init__(self, cli=None):
        super(VNXCliResource, self).__init__()
        self.poll = True
        self._cli = cli
        self.timestamp = None

    def shadow_copy(self):
        ret = super(VNXCliResource, self).shadow_copy()
        ret._cli = self._cli
        ret.poll = self.poll
        ret.timestamp = self.timestamp
        return ret

    def with_poll(self):
        ret = _WithPoll(self)
        self.poll = True
        return ret

    def with_no_poll(self):
        ret = _WithPoll(self)
        self.poll = False
        return ret

    @instance_cache
    def _get_resource_property(self, value):
        value.set_cli(self._cli)
        return value

    def set_cli(self, cli):
        if cli is not None:
            self._cli = cli

    @clear_instance_cache
    def update(self, data=None):
        ret = super(VNXCliResource, self).update(data)
        self.timestamp = datetime.now()
        return ret

    def _get_property_from_raw(self, item):
        if item in self.metric_names():
            value = self.get_metric_value(item)
        else:
            value = super(VNXCliResource, self)._get_property_from_raw(item)
            if isinstance(value, VNXCliResource):
                value = self._get_resource_property(value)
        return value

    def property_names(self):
        names = super(VNXCliResource, self).property_names()
        if self._cli is not None and self._cli.is_perf_metric_enabled(self):
            names.extend(self.metric_names())
        return names

    def metric_names(self):
        return calculators.get_metric_names(self.resource_class_name())

    def get_metric_value(self, item):
        if not self._cli.is_perf_metric_enabled(self):
            raise VNXPerMonNotEnabledError()
        return calculators.get_metric_value(
            self.resource_class_name(), item, self._cli, self)


class WithListPoll(object):
    def __init__(self, rsc_list):
        self._rsc_list = rsc_list
        self._orig_polls = {rsc: rsc.poll for rsc in self._rsc_list}

    def __enter__(self):
        pass

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type, exc_val, exc_tb):
        # return None, do not handle inner exception
        for rsc in self._rsc_list:
            rsc.poll = self._orig_polls[rsc]


def _hdr_cb(rsc):
    if hasattr(rsc, 'name'):
        name = rsc.name
    elif hasattr(rsc, 'index'):
        name = rsc.index
    else:
        raise AttributeError('resource should have "name" or "index" defined.')

    return [rsc.timestamp.isoformat(str(' ')), str(name)]


class VNXCliResourceList(VNXCliResource, ResourceList):
    def __init__(self, cli=None):
        VNXCliResource.__init__(self, cli=cli)
        ResourceList.__init__(self)

        extra_headers = ['timestamp', 'name']
        self._metrics_dumper = MetricsDumper(
            self, extra_headers, _hdr_cb)

        self._poll = True

    def shadow_copy(self, *args, **kwargs):
        ret = VNXCliResource.shadow_copy(self)
        ret.set_filter(*args, **kwargs)
        return ret

    @classmethod
    def _get_parser(cls):
        return get_vnx_parser(cls.get_resource_class().__name__)

    @classmethod
    def get_resource_class(cls):
        raise NotImplementedError(
            'should return the class ref of the resource in the list.')

    @classmethod
    def resource_class_name(cls):
        return cls.get_resource_class().resource_class_name()

    @property
    def poll(self):
        return self._poll

    @poll.setter
    def poll(self, value):
        self._poll = value
        if self._is_updated():
            for item in self:
                item.poll = self._poll

    def _get_resource_instance(self):
        clz = self.get_resource_class()
        if issubclass(clz, VNXCliResource):
            ret = clz(cli=self._cli)
            ret.poll = self.poll
        else:
            ret = clz()
        return ret

    @clear_instance_cache
    def update(self, data=None):
        ret = super(VNXCliResourceList, self).update(data)
        for item in self._list:
            item._cli = self._cli
            item.poll = self.poll
        return ret

    def set_cli(self, cli):
        super(VNXCliResourceList, self).set_cli(cli)
        for item in self:
            if isinstance(item, VNXCliResource):
                item.set_cli(cli)

    def persist_metric_data(self, filename=None):
        if filename is None:
            filename = self.get_default_metric_csv_filename()
        return self._metrics_dumper.persist_metric_data(filename)

    def get_default_metric_csv_filename(self):
        folder = get_local_folder()
        name = '{}_{}.csv'.format(self._cli.ip, self.resource_class_name())
        return os.path.join(folder, name)

    def get_metrics_csv(self, sep=None):
        return self._metrics_dumper.get_metrics_csv(sep=sep)



# Copyright (c) 2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, instance_of, has_items, raises, \
    contains_string, none, calling

from storops.exception import UnityStorageResourceNameInUseError, \
    UnityConsistencyGroupNameInUseError, UnityResourceNotFoundError, \
    UnityNothingToModifyError, UnityHostAccessAlreadyExistsError
from storops.unity.enums import StorageResourceTypeEnum, HostLUNAccessEnum
from storops.unity.resource.cg import UnityConsistencyGroup, \
    UnityConsistencyGroupList

from storops.unity.resource.host import UnityHost
from storops.unity.resource.lun import UnityLun
from storops.unity.resource.remote_system import UnityRemoteSystem
from storops.unity.resource.snap import UnitySnap

from storops_test.unity.rest_mock import t_rest, patch_rest


__author__ = 'Peter Wang'


class UnityConsistencyGroupTest(TestCase):
    cg_type = StorageResourceTypeEnum.CONSISTENCY_GROUP

    @staticmethod
    def get_cg():
        return UnityConsistencyGroup(cli=t_rest(), _id='res_19')

    @patch_rest
    def test_get_properties(self):
        cg = UnityConsistencyGroup(_id='res_13', cli=t_rest())
        assert_that(cg.id, equal_to('res_13'))
        assert_that(cg.type, equal_to(self.cg_type))

    @patch_rest
    def test_get_cg_list(self):
        cg_list = UnityConsistencyGroupList(cli=t_rest())
        assert_that(len(cg_list), equal_to(2))
        for cg in cg_list:
            assert_that(cg, instance_of(UnityConsistencyGroup))

    @patch_rest
    def test_create_empty_cg(self):
        cg = UnityConsistencyGroup.create(t_rest(), 'Goddess')
        assert_that(cg.name, equal_to('Goddess'))
        assert_that(cg.type, equal_to(self.cg_type))

    @patch_rest
    def test_create_cg_with_initial_member(self):
        lun1 = UnityLun(cli=t_rest(), _id='sv_3339')
        lun2 = UnityLun(cli=t_rest(), _id='sv_3340')
        cg = UnityConsistencyGroup.create(t_rest(), 'Muse',
                                          lun_add=[lun1, lun2])
        assert_that(cg.name, equal_to('Muse'))

        members = cg.luns
        assert_that(len(members), equal_to(2))
        lun_id_list = map(lambda lun: lun.get_id(), members)
        assert_that(lun_id_list, has_items('sv_3339', 'sv_3340'))

    @patch_rest
    def test_create_cg_with_hosts(self):
        lun1 = UnityLun(cli=t_rest(), _id='sv_3338')
        host1 = UnityHost(cli=t_rest(), _id='Host_14')
        host2 = UnityHost(cli=t_rest(), _id='Host_15')
        cg = UnityConsistencyGroup.create(
            t_rest(), 'Muse', lun_add=[lun1], hosts=[host1, host2])
        hosts = cg.block_host_access

        assert_that(len(hosts), equal_to(2))
        for mask in hosts.access_mask:
            assert_that(mask, equal_to(HostLUNAccessEnum.BOTH))

    @patch_rest
    def test_create_cg_name_in_use(self):
        def f():
            UnityConsistencyGroup.create(t_rest(), 'in_use')

        assert_that(f, raises(UnityConsistencyGroupNameInUseError, 'used'))

    @patch_rest
    def test_delete_cg_normal(self):
        cg = UnityConsistencyGroup(cli=t_rest(), _id='res_18')
        resp = cg.delete()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete_cg_not_exists(self):
        def f():
            cg = UnityConsistencyGroup(cli=t_rest(), _id='res_119')
            cg.delete()

        assert_that(f, raises(UnityResourceNotFoundError, 'not exist'))

    @patch_rest
    def test_rename_name_used(self):
        def f():
            self.get_cg().name = 'iscsi-test'

        assert_that(f, raises(UnityStorageResourceNameInUseError, 'reserved'))

    @patch_rest
    def test_add_lun_success(self):
        lun1 = UnityLun(cli=t_rest(), _id='sv_3341')
        lun2 = UnityLun(cli=t_rest(), _id='sv_3342')
        resp = self.get_cg().add_lun(lun1, lun2)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_add_lun_nothing_to_modify(self):
        def f():
            lun = UnityLun(cli=t_rest(), _id='sv_3341')
            self.get_cg().add_lun(lun)

        assert_that(f, raises(UnityNothingToModifyError, 'nothing to modify'))

    @patch_rest
    def test_remove_lun_success(self):
        lun = UnityLun(cli=t_rest(), _id='sv_3342')
        resp = self.get_cg().remove_lun(lun)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_remove_host_access(self):
        host1 = UnityHost(cli=t_rest(), _id='Host_14')
        host2 = UnityHost(cli=t_rest(), _id='Host_15')
        resp = self.get_cg().remove_host_access(host1, host2)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_set_host_access(self):
        host = UnityHost(cli=t_rest(), _id='Host_14')
        resp = self.get_cg().set_host_access(host)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_add_host_access_existed(self):
        def f():
            host1 = UnityHost(cli=t_rest(), _id='Host_14')
            host2 = UnityHost(cli=t_rest(), _id='Host_15')
            self.get_cg().add_host_access(host1, host2)

        assert_that(f, raises(UnityHostAccessAlreadyExistsError, 'has access'))

    @patch_rest
    def test_add_host_access_success(self):
        host1 = UnityHost(cli=t_rest(), _id='Host_12')
        host2 = UnityHost(cli=t_rest(), _id='Host_15')
        resp = self.get_cg().add_host_access(host1, host2)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_create_cg_snap(self):
        snap = self.get_cg().create_snap('song')
        assert_that(snap.name, equal_to('song'))
        assert_that(snap.storage_resource.get_id(), equal_to('res_19'))

    @patch_rest
    def test_list_cg_snaps(self):
        snaps = self.get_cg().snapshots
        assert_that(len(snaps), equal_to(2))
        assert_that(map(lambda s: s.name, snaps), has_items('song', 'tragedy'))

    @patch_rest
    def test_detach_cg_snap(self):
        snap = UnitySnap(cli=t_rest(), _id='85899345927')
        host = UnityHost(cli=t_rest(), _id='Host_22')
        resp = snap.detach_from(host)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_attach_cg_snap(self):
        snap = UnitySnap(cli=t_rest(), _id='85899345927')
        host = UnityHost(cli=t_rest(), _id='Host_22')
        resp = snap.attach_to(host)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_restore_cg_snap_with_backup(self):
        snap = UnitySnap(cli=t_rest(), _id='85899345927')
        backup_snap = snap.restore('paint')
        assert_that(backup_snap.name, equal_to('paint'))
        assert_that(backup_snap.storage_resource.get_id(), equal_to('res_19'))

    @patch_rest
    def test_restore_cg_snap_without_backup(self):
        snap = UnitySnap(cli=t_rest(), _id='85899345927')
        backup_snap = snap.restore()
        assert_that(backup_snap.name, equal_to('2016-11-03_08.35.00'))
        assert_that(backup_snap.storage_resource.get_id(), equal_to('res_19'))

    @patch_rest
    def test_attach_lun_in_cg(self):
        lun_in_cg = UnityLun.get(cli=t_rest(), _id='sv_15')
        host = UnityHost(cli=t_rest(), _id='Host_22')

        resp = lun_in_cg.attach_to(host)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_detach_lun_in_cg(self):
        lun_in_cg = UnityLun.get(cli=t_rest(), _id='sv_15')
        host = UnityHost(cli=t_rest(), _id='Host_2')

        resp = lun_in_cg.detach_from(host)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_detach_lun_nothing(self):
        lun_in_cg = UnityLun.get(cli=t_rest(), _id='sv_3')
        host = UnityHost(cli=t_rest(), _id='Host_2')

        r = lun_in_cg.detach_from(host)
        assert_that(r, none())

    @patch_rest
    def test_detach_lun_from_all(self):
        lun_in_cg = UnityLun.get(cli=t_rest(), _id='sv_15')
        resp = lun_in_cg.detach_from(None)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_expand_lun_in_cg(self):
        lun_in_cg = UnityLun.get(cli=t_rest(), _id='sv_15')
        old_size = lun_in_cg.size_total
        resp = lun_in_cg.expand(100 * 1024 ** 3)
        assert_that(resp, equal_to(old_size))

    @patch_rest
    def test_modify_lun_in_cg(self):
        lun_in_cg = UnityLun.get(cli=t_rest(), _id='sv_15')
        resp = lun_in_cg.modify(size=15 * 1024 ** 3)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_modify_lun_compression_enabled_in_cg_v4_2(self):
        lun_in_cg = UnityLun.get(cli=t_rest(version='4.2'), _id='sv_17')
        resp = lun_in_cg.modify(is_compression=True)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_modify_lun_compression_enabled_in_cg(self):
        lun_in_cg = UnityLun.get(cli=t_rest(version='4.3'), _id='sv_18')
        resp = lun_in_cg.modify(is_compression=True)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_replace_lun_in_cg(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_6')
        lun1 = UnityLun(cli=t_rest(), _id='sv_3341')
        lun2 = UnityLun(cli=t_rest(), _id='sv_3342')
        resp = cg.replace_lun(lun1, lun2)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_replace_lun_in_cg_already_in(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_6')
        lun1 = UnityLun(cli=t_rest(), _id='sv_3341')
        lun2 = UnityLun(cli=t_rest(), _id='sv_14')
        resp = cg.replace_lun(lun1, lun2)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_get_csv(self):
        cg_list = UnityConsistencyGroupList(cli=t_rest())
        csv = cg_list.get_metrics_csv()
        assert_that(csv, contains_string('id,name'))
        assert_that(csv, contains_string('res_3,smis-test-cg'))

    @patch_rest
    def test_attach_cg_not_implemented(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_6')
        assert_that(calling(cg.attach_to), raises(NotImplementedError))

    @patch_rest
    def test_detach_cg_not_implemented(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_6')
        assert_that(calling(cg.detach_from), raises(NotImplementedError))

    @patch_rest
    def test_update_lun_add(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_21')
        lun1 = UnityLun(cli=t_rest(), _id='sv_60')
        lun2 = UnityLun(cli=t_rest(), _id='sv_61')
        resp = cg.update_lun(add_luns=[lun1, lun2])
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_update_lun_remove(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_21')
        lun1 = UnityLun(cli=t_rest(), _id='sv_58')
        lun2 = UnityLun(cli=t_rest(), _id='sv_59')
        resp = cg.update_lun(remove_luns=[lun1, lun2])
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_update_lun_add_remove(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_21')
        lun1 = UnityLun(cli=t_rest(), _id='sv_58')
        lun2 = UnityLun(cli=t_rest(), _id='sv_59')
        lun3 = UnityLun(cli=t_rest(), _id='sv_60')
        lun4 = UnityLun(cli=t_rest(), _id='sv_61')
        resp = cg.update_lun(add_luns=[lun1, lun3], remove_luns=[lun2, lun4])
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_update_lun_all_none(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_21')
        resp = cg.update_lun()
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_cg_is_replicated(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_6')
        result = cg.check_cg_is_replicated()
        assert_that(result, equal_to(True))

    @patch_rest
    def test_cg_is_not_replicated(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_21')
        result = cg.check_cg_is_replicated()
        assert_that(result, equal_to(None))

    @patch_rest
    def test_create_replicate_cg_session(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_21')
        rs = UnityRemoteSystem.get(cli=t_rest(), _id='RS_0')
        lun1 = UnityLun(cli=t_rest(), _id='sv_58')
        lun2 = UnityLun(cli=t_rest(), _id='sv_59')
        luns = [lun1, lun2]
        rep_session = cg.replicate_cg_with_dst_resource_provisioning(
            60, luns, 'pool_2', dst_cg_name=cg.name, remote_system=rs)
        assert_that(rep_session.id, equal_to(
            '81866043753_A9221PM0010744_0604000_81378689_FNM84900101113_0000'))

    @patch_rest
    def test_create_local_replicate_cg_session(self):
        # Create the session in local when the remote system is empty
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_21')
        lun1 = UnityLun(cli=t_rest(), _id='sv_58')
        lun2 = UnityLun(cli=t_rest(), _id='sv_59')
        luns = [lun1, lun2]
        rep_session = cg.replicate_cg_with_dst_resource_provisioning(
            60, luns, 'pool_2', dst_cg_name=cg.name, remote_system=None)
        assert_that(rep_session.id, equal_to(
            '81604378625_APM00192210744_0000_81604378629_APM00192210744_0000'))

    @patch_rest
    def test_create_replicate_cg_session_without_dst_cg(self):
        cg = UnityConsistencyGroup.get(cli=t_rest(), _id='res_21')
        rs = UnityRemoteSystem.get(cli=t_rest(), _id='RS_0')
        lun1 = UnityLun(cli=t_rest(), _id='sv_58')
        lun2 = UnityLun(cli=t_rest(), _id='sv_59')
        luns = [lun1, lun2]
        rep_session = cg.replicate_cg_with_dst_resource_provisioning(
            60, luns, 'pool_2', dst_cg_name=None, remote_system=rs)
        assert_that(rep_session.id, equal_to(
            '81866043753_A9221PM0010744_0604000_81378689_FNM84900101113_0000'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import io
import os
import re

from setuptools import setup, find_packages

__author__ = 'Cedric Zhuang'


def version():
    desc = get_long_description()
    ret = re.findall(r'VERSION: (.*)', desc)[0]
    return ret.strip()


def here(filename=None):
    ret = os.path.abspath(os.path.dirname(__file__))
    if filename is not None:
        ret = os.path.join(ret, filename)
    return ret


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n\n')
    buf = []
    for filename in filenames:
        with io.open(here(filename), encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


def read_requirements(filename):
    with open(filename) as f:
        return f.read().splitlines()


def get_description():
    return 'Python API for VNX and Unity.'


def get_long_description():
    filename = 'README.rst'
    return read(filename)


setup(
    name='storops',
    version=version(),
    author='Cedric Zhuang',
    author_email='cedric.zhuang@gmail.com',
    maintainer='Dell EMC Unity OpenStack Team',
    maintainer_email='spe.unity.openstack@emc.com',
    url='https://github.com/emc-openstack/storops',
    description=get_description(),
    license='Apache Software License',
    keywords=['VNX', 'Unity', 'EMC', 'Storage'],
    include_package_data=True,
    packages=find_packages(),
    platforms=['any'],
    long_description=get_long_description(),
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Natural Language :: English',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
        'License :: OSI Approved :: Apache Software License',
    ],
    install_requires=read_requirements('requirements.txt'),
    tests_require=read_requirements('test-requirements.txt')
)



# coding=utf-8
# Copyright (c) 2017 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from contextlib import contextmanager

from storops.vnx.resource import VNXCliResourceList, VNXCliResource
from storops.vnx.resource.lun import VNXLun, VNXLunList
from storops.vnx.enums import VNXCtrlMethod
from storops import exception as ex

__author__ = "Peter Wang"


def convert_lun(luns):
    lun_ids = []
    smp_names = []
    for lun in luns:
        if lun.is_snap_mount_point:
            smp_names.append(lun.name)
        else:
            lun_ids.append(lun.lun_id)
    return lun_ids, smp_names


def normalize_lun(luns, cli):
    if isinstance(luns, int):
        return [VNXLun(lun_id=luns, cli=cli)]
    elif isinstance(luns, VNXLun):
        return [luns]
    elif isinstance(luns, list) or isinstance(luns, VNXLunList):
        return luns
    elif luns is None:
        return []
    else:
        raise ValueError('Invalid format for luns.')


def convert_ioclass(ioclasses):
    names = []
    if ioclasses:
        for ioclass in ioclasses:
            names.append(ioclass.name)
    return names


@contextmanager
def restart_policy(policy):
    VNXIOPolicy.stop_policy(cli=policy._cli)
    try:
        yield policy
    except ex.StoropsException:
        pass
    policy.run_policy()


class VNXIOClassList(VNXCliResourceList):
    def __init__(self, cli=None, name=None):
        super(VNXIOClassList, self).__init__()
        self._cli = cli
        self._name = name

    @classmethod
    def get_resource_class(cls):
        return VNXIOClass

    def _get_raw_resource(self):
        return self._cli.get_ioclass(name=self._name, poll=self.poll)

    def _filter(self, item):
        """Filter the system default class"""
        return 'Background Class' not in item.name


class VNXIOClass(VNXCliResource):
    def __init__(self, name=None, cli=None):
        super(VNXIOClass, self).__init__()
        self._cli = cli
        self._name = name

    def _get_raw_resource(self):
        if self._cli is None:
            raise ValueError('client is not available for this resource.')
        return self._cli.get_ioclass(name=self._name, poll=self.poll)

    @staticmethod
    def get(cli, name=None):
        ret = VNXIOClassList(cli=cli)
        if name:
            ret = VNXIOClass(name=name, cli=cli)
        return ret

    @property
    def luns(self):
        """Aggregator for ioclass_luns and ioclass_snapshots."""
        lun_list, smp_list = [], []
        if self.ioclass_luns:
            lun_list = map(lambda l: VNXLun(lun_id=l.lun_id, name=l.name,
                                            cli=self._cli), self.ioclass_luns)
        if self.ioclass_snapshots:
            smp_list = map(lambda smp: VNXLun(name=smp.name, cli=self._cli),
                           self.ioclass_snapshots)
        return list(lun_list) + list(smp_list)

    @property
    def policy(self):
        """Returns policy which contains this ioclass."""
        policies = VNXIOPolicy.get(cli=self._cli)
        ret = None
        for policy in policies:
            contained = policy.ioclasses.name
            if self._get_name() in contained:
                ret = VNXIOPolicy.get(name=policy.name, cli=self._cli)
                break
        return ret

    @staticmethod
    def create(cli, name, iotype=None, luns=None,
               ctrlmethod=VNXCtrlMethod.NO_CTRL, minsize=None, maxsize=None):
        luns = normalize_lun(luns, cli)
        lun_ids, smp_names = [], []
        if luns:
            lun_ids, smp_names = convert_lun(luns)
        out = cli.create_ioclass(name, iotype, lun_ids, smp_names,
                                 ctrlmethod=ctrlmethod, minsize=minsize,
                                 maxsize=maxsize)
        ex.raise_if_err(out, default=ex.VNXIOClassError)
        return VNXIOClass(name, cli)

    def modify(self, new_name=None, iotype=None, lun_ids=None, smp_names=None,
               ctrlmethod=None, minsize=None, maxsize=None):
        """Overwrite the current properties for a VNX ioclass.

        :param new_name: new name for the ioclass
        :param iotype: can be 'rw', 'r' or 'w'
        :param lun_ids: list of LUN IDs
        :param smp_names: list of Snapshot Mount Point names
        :param ctrlmethod: the new CtrlMethod
        :param minsize: minimal size in kb
        :param maxsize: maximum size in kb
        """
        if not any([new_name, iotype, lun_ids, smp_names, ctrlmethod]):
            raise ValueError('Cannot apply modification, please specify '
                             'parameters to modify.')

        def _do_modify():
            out = self._cli.modify_ioclass(
                self._get_name(), new_name, iotype, lun_ids, smp_names,
                ctrlmethod, minsize, maxsize)
            ex.raise_if_err(out, default=ex.VNXIOClassError)

        try:
            _do_modify()
        except ex.VNXIOCLassRunningError:
            with restart_policy(self.policy):
                _do_modify()

        return VNXIOClass(new_name if new_name else self._get_name(),
                          self._cli)

    def add_lun(self, luns):
        """A wrapper for modify method.

        .. note:: This API only append luns to existing luns.
        """
        curr_lun_ids, curr_smp_names = self._get_current_names()
        luns = normalize_lun(luns, self._cli)
        new_ids, new_smps = convert_lun(luns)
        if new_ids:
            curr_lun_ids.extend(new_ids)
        if new_smps:
            curr_smp_names.extend(new_smps)
        return self.modify(lun_ids=curr_lun_ids, smp_names=curr_smp_names)

    def remove_lun(self, luns):
        luns = normalize_lun(luns, self._cli)
        curr_lun_ids, curr_smp_names = self._get_current_names()
        rm_ids, rm_smps = convert_lun(luns)
        new_ids = [curr for curr in curr_lun_ids if curr not in rm_ids]
        new_smps = [curr for curr in curr_smp_names if curr not in rm_smps]
        return self.modify(lun_ids=new_ids, smp_names=new_smps)

    def delete(self):
        out = self._cli.delete_ioclass(self._get_name())
        ex.raise_if_err(out, default=ex.VNXIOClassError)
        return VNXIOClass.get(name=self._get_name(), cli=self._cli)

    def add_to_policy(self, policy):
        if not isinstance(policy, VNXIOPolicy):
            policy = VNXIOPolicy(self._cli, policy)
        return policy.add_class(self)

    def _get_current_names(self):
        lun_ids = []
        smp_names = []

        for lun in self.ioclass_luns:
            lun_ids.append(lun.lun_id)
        for smp in self.ioclass_snapshots:
            smp_names.append(smp.name)

        return lun_ids, smp_names


class VNXIOPolicyList(VNXCliResourceList):
    def __init__(self, cli=None, name=None):
        super(VNXIOPolicyList, self).__init__()
        self._cli = cli
        self._name = name

    @classmethod
    def get_resource_class(cls):
        return VNXIOPolicy

    def _get_raw_resource(self):
        return self._cli.get_policy(name=self._name, poll=self.poll)


class VNXIOPolicy(VNXCliResource):
    def __init__(self, name=None, cli=None):
        super(VNXIOPolicy, self).__init__()
        self._name = name
        self._cli = cli

    def _get_raw_resource(self):
        if self._cli is None:
            raise ValueError('client is not available for this resource.')
        return self._cli.get_policy(name=self._name, poll=self.poll)

    @staticmethod
    def get(cli, name=None):
        ret = VNXIOPolicyList(cli=cli)
        if name:
            ret = VNXIOPolicy(cli=cli, name=name)
        return ret

    @staticmethod
    def create(cli, name, ioclasses=None, fail_action=None, time_limit=None,
               eval_window=None):
        ioclass_names = convert_ioclass(ioclasses)
        out = cli.create_policy(name, ioclass_names, fail_action, time_limit,
                                eval_window)
        ex.raise_if_err(out, default=ex.VNXIOPolicyError)
        return VNXIOPolicy(name, cli)

    def modify(self, new_name=None, new_ioclasses=None, time_limit=None,
               fail_action=None, eval_window=None):
        new_names = convert_ioclass(new_ioclasses)

        def _do_modify():
            out = self._cli.modify_policy(
                self._get_name(), new_name, new_names, time_limit,
                fail_action, eval_window)
            ex.raise_if_err(out, default=ex.VNXIOPolicyError)

        try:
            _do_modify()
        except ex.VNXIOPolicyRunningError:
            with restart_policy(self):
                _do_modify()

    def delete(self):
        out = self._cli.delete_policy(self._get_name())
        ex.raise_if_err(out)
        return VNXIOPolicy.get(name=self._get_name(), cli=self._cli)

    def add_class(self, ioclass):
        """Add one VNXIOClass instance to policy.

        .. note: due to the limitation of VNX, need to stop the policy first.
        """
        current_ioclasses = self.ioclasses
        if ioclass.name in current_ioclasses.name:
            return
        current_ioclasses.append(ioclass)
        self.modify(new_ioclasses=current_ioclasses)

    def remove_class(self, ioclass):
        """Remove VNXIOClass instance from policy."""
        current_ioclasses = self.ioclasses
        new_ioclasses = filter(lambda x: x.name != ioclass.name,
                               current_ioclasses)
        self.modify(new_ioclasses=new_ioclasses)

    def run_policy(self):
        out = self._cli.run_policy(self._get_name())
        ex.raise_if_err(out, default=ex.VNXIOPolicyError)

    @staticmethod
    def stop_policy(cli):
        out = cli.stop_policy()
        ex.raise_if_err(out, default=ex.VNXIOPolicyError)

    def measure_policy(self):
        out = self._cli.measure_policy(self._get_name())
        ex.raise_if_err(out, default=ex.VNXIOPolicyError)


class VNXIOClassLunList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXIOClassLun

    def _filter(self, item):
        """Filter the wierd outpout.

        :returns: False if contains 'Not Lun Specific'
        """
        return item.existed


class VNXIOClassLun(VNXCliResource):
    pass


class VNXIOClassSnapshot(VNXCliResource):
    pass


class VNXIOClassSnapshotList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXIOClassSnapshot



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging

from retryz import retry

from storops_comptest import t_vnx, vnx1, vnx2
from storops_comptest.utils import ResourceManager
from storops import exception as ex

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class VNXTestResourceManager(ResourceManager):
    def __init__(self, name):
        super(VNXTestResourceManager, self).__init__(name)
        self.vnx = None
        self.pool = None
        self.lun = None

    def _create_cg(self, name):
        if not self.has_cg_name(name):
            try:
                ret = self.vnx.create_cg(name)
            except ex.VNXConsistencyGroupNameInUseError:
                ret = self.vnx.get_cg(name)
            self.add_cg_name(name)
        else:
            ret = self.vnx.get_cg(name)
        return ret

    def _create_sg(self, name):
        if not self.has_sg_name(name):
            try:
                ret = self.vnx.create_sg(name)
            except ex.VNXStorageGroupNameInUseError:
                ret = self.vnx.get_sg(name=name)
            self.add_sg_name(name)
        else:
            ret = self.vnx.get_sg(name=name)
        ret.shuffle_hlu = False
        return ret

    def _create_snap(self, name):
        if not self.has_snap_name(name):

            @retry(on_error=ex.VNXLunPreparingError, wait=7)
            def _do_create():
                try:
                    snap = self.lun.create_snap(name, allow_rw=True)
                except ex.VNXSnapNameInUseError:
                    snap = self.vnx.get_snap(name)
                self.add_snap_name(name)
                return snap

            ret = _do_create()
        else:
            ret = self.vnx.get_snap(name)
        return ret

    def _get_pool(self, vnx=None):
        if vnx is None:
            vnx = self.vnx
        pools = vnx.get_pool()
        if pools:
            ret = pools[0]
        else:
            ret = None
        return ret

    def _create_pool(self, name, vnx=None):
        if vnx is None:
            vnx = self.vnx
        if not self.has_pool_name(name):

            @retry(on_error=ex.VNXDiskUsedError, wait=7)
            def _do_create():
                try:
                    pool = vnx.create_pool(name)
                except ex.VNXPoolNameInUseError:
                    pool = vnx.get_pool(name=name)
                self.add_pool_name(name)
                return pool

            ret = _do_create()
        else:
            ret = vnx.get_pool(name=name)
        return ret

    def _create_lun(self, name, vnx=None, pool=None):
        if vnx is None:
            vnx = self.vnx
        if pool is None:
            pool = self.pool

        if not self.has_lun_name(name):
            try:
                ret = pool.create_lun(name)
            except ex.VNXLunNameInUseError:
                ret = vnx.get_lun(name=name)
            self.add_lun_name(name)
        else:
            ret = vnx.get_lun(name=name)
        return ret

    def clean_up(self):
        super(VNXTestResourceManager, self).clean_up()

        self._clean_up_snap()
        self._clean_up_cg()
        self._clean_up_mirror()
        self._clean_up_sg()
        self._clean_up_lun()
        self._clean_up_pool()
        self._names.destroy()

    def _clean_up_pool(self):
        while self.has_name('pool'):
            name = None
            try:
                name = self._pop_name('pool')
                pool = self.vnx.get_pool(name=name)
                if pool.existed:
                    pool.delete(force=True)
            except (ex.VNXPoolNotFoundError, ex.VNXPoolDestroyingError,
                    IndexError):
                log.exception('delete pool {} failed.'.format(name))

    def _clean_up_lun(self, vnx=None):
        if vnx is None:
            vnx = self.vnx
        while self.has_name('lun'):
            name = None
            try:
                name = self._pop_name('lun')
                lun = vnx.get_lun(name=name)
                if lun.existed:
                    lun.delete(force=True)
            except (ex.VNXLunNotFoundError, ex.VNXDeleteLunError, IndexError):
                log.exception('delete lun {} failed.'.format(name))

    def _clean_up_sg(self):
        while self.has_name('sg'):
            name = None
            try:
                name = self._pop_name('sg')
                sg = self.vnx.get_sg(name=name)
                if sg.existed:
                    sg.delete(disconnect_host=True)
            except (ex.VNXStorageGroupNotFoundError, IndexError):
                log.exception('delete sg {} failed.'.format(name))

    def _clean_up_cg(self):
        while self.has_name('cg'):
            name = None
            try:
                name = self._pop_name('cg')
                cg = self.vnx.get_cg(name=name)
                if cg.existed:
                    cg.delete()
            except (ex.VNXConsistencyGroupNotFoundError, IndexError):
                log.exception('delete cg {} failed.'.format(name))

    def _clean_up_mirror(self):
        @retry(on_error=ex.VNXMirrorRemoveSynchronizingError, wait=10,
               limit=18)
        def _remove_image(m):
            m.fracture_image()
            m.remove_image()

        while self.has_name('mirror'):
            name = None
            try:
                name = self._pop_name('mirror')
                mirror = self.vnx.get_mirror_view(name=name)
                if mirror.existed:
                    if len(mirror.images) > 1:
                        _remove_image(mirror)
                    mirror.delete()
            except ex.VNXMirrorException:
                log.exception('delete mirror {} failed.'.format(name))

    def _clean_up_snap(self):
        while self.has_snap_name():
            self._pop_name('snap')


class VNXGeneralFixtureManager(VNXTestResourceManager):
    def __init__(self):
        super(VNXGeneralFixtureManager, self).__init__('vg')
        self.cg = None
        self.snap = None
        self.sg = None

    def setup(self):
        clz_name = self.__class__.__name__
        log.debug('start {} setup.'.format(clz_name))
        try:
            super(VNXGeneralFixtureManager, self).setup()
            self.vnx = t_vnx()
            self.cg = self._create_cg('general_cg')
            self.pool = self._get_pool()
            if self.pool is None:
                self.pool = self._create_pool('general_pool')
            self.lun = self._create_lun('general_lun')
            self.snap = self._create_snap('general_snap')
            self.sg = self._create_sg('general_sg')
        except ex.StoropsException:
            log.exception('failed to initialize {}'.format(clz_name))
            raise

    def select_port(self):
        ports = self.vnx.get_iscsi_port()
        for port in ports:
            if port.ip_address is None:
                break
        else:
            raise ValueError('no available iSCSI port found.')
        return port.sp, port.port_id


class MultiVNXGeneralFixtureManager(VNXTestResourceManager):
    src_lun = 'general_mirror_src'
    tgt_lun = 'general_mirror_tgt'

    def __init__(self):
        super(MultiVNXGeneralFixtureManager, self).__init__('mvg')
        self.vnx1 = None
        self.vnx2 = None
        self.lun1 = None
        self.lun2 = None
        self.mirror = None

    def setup(self):
        clz_name = self.__class__.__name__
        log.debug('start {} setup.'.format(clz_name))
        try:
            super(MultiVNXGeneralFixtureManager, self).setup()
            self.vnx = vnx1()
            self.vnx1 = self.vnx
            self.vnx2 = vnx2()
            self.lun = self._get_mirror_lun(self.vnx1, self.src_lun)
            self.lun1 = self.lun
            self.lun2 = self._get_mirror_lun(self.vnx2, self.tgt_lun)
            self.mirror = self._create_mirror('general_sync_mirror')
        except ex.StoropsException:
            log.exception('failed to initialize {}'.format(clz_name))
            raise

    def _get_mirror_lun(self, vnx, name):
        pool = self._get_pool(vnx=vnx)
        if pool is None:
            pool = self._create_pool('general_pool', vnx=vnx)
        return self._create_lun(name, vnx=vnx, pool=pool)

    def _create_mirror(self, name):
        if not self.has_mirror_name(name):
            try:
                ret = self.lun1.create_mirror_view(name, False)
            except ex.VNXMirrorNameInUseError:
                ret = self.vnx1.get_mirror_view(name)
            self.add_mirror_name(name)
        else:
            ret = self.vnx.get_mirror_view(name)

        ret.add_image(self.vnx2.spa_ip, self.lun2.lun_id)
        return ret

    def clean_up(self):
        super(MultiVNXGeneralFixtureManager, self).clean_up()

        # remote the remote LUN
        try:
            lun = self.vnx2.get_lun(name=self.tgt_lun)
            if lun.existed:
                lun.delete(force=True)
        except (ex.VNXLunNotFoundError, ex.VNXDeleteLunError, IndexError):
            log.exception('delete lun {} failed.'.format(self.tgt_lun))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to

from storops_test.vnx.cli_mock import patch_cli, t_cli
from storops.vnx.resource.ndu import VNXNdu

__author__ = 'Cedric Zhuang'


class VNXNduTest(TestCase):
    @patch_cli
    def test_get_all(self):
        ndu_list = VNXNdu.get(t_cli())
        assert_that(len(ndu_list), equal_to(16))

    @patch_cli
    def test_get(self):
        ndu = VNXNdu.get(t_cli(), '-VNXSnapshots')
        assert_that(ndu.name, equal_to('-VNXSnapshots'))
        assert_that(ndu.revision, equal_to('-'))
        assert_that(ndu.commit_required, equal_to(False))
        assert_that(ndu.revert_possible, equal_to(False))
        assert_that(ndu.active_state, equal_to(True))
        assert_that(ndu.is_installation_completed, equal_to(True))
        assert_that(ndu.is_this_system_software, equal_to(False))

    @patch_cli(output='-np_ndu_-list_-name_-Deduplication_no.txt')
    def test_is_dedup_enabled_false(self):
        assert_that(VNXNdu.is_dedup_enabled(t_cli()), equal_to(False))

    @patch_cli
    def test_is_dedup_enabled(self):
        assert_that(VNXNdu.is_dedup_enabled(t_cli()), equal_to(True))

    @patch_cli
    def test_is_compression_enabled(self):
        assert_that(VNXNdu.is_compression_enabled(t_cli()), equal_to(True))

    @patch_cli
    def test_is_snap_enabled(self):
        assert_that(VNXNdu.is_snap_enabled(t_cli()), equal_to(True))

    @patch_cli
    def test_is_mirror_view_async_enabled(self):
        assert_that(VNXNdu.is_mirror_view_async_enabled(t_cli()),
                    equal_to(True))

    @patch_cli
    def test_is_mirror_view_sync_enabled(self):
        assert_that(VNXNdu.is_mirror_view_sync_enabled(t_cli()),
                    equal_to(True))

    @patch_cli
    def test_is_mirror_view_enabled(self):
        assert_that(VNXNdu.is_mirror_view_enabled(t_cli()),
                    equal_to(True))

    @patch_cli
    def test_is_thin_enabled(self):
        assert_that(VNXNdu.is_thin_enabled(t_cli()), equal_to(True))

    @patch_cli
    def test_is_sancopy_enabled(self):
        assert_that(VNXNdu.is_sancopy_enabled(t_cli()), equal_to(True))

    @patch_cli
    def test_is_auto_tiering_enabled(self):
        assert_that(VNXNdu.is_auto_tiering_enabled(t_cli()), equal_to(True))

    @patch_cli
    def test_is_fast_cache_enabled(self):
        assert_that(VNXNdu.is_fast_cache_enabled(t_cli()), equal_to(True))



# coding=utf-8
# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import unicode_literals

from unittest import TestCase

import ddt
from hamcrest import assert_that, equal_to, raises, calling

from storops.unity.enums import FilesystemSnapAccessTypeEnum, \
    ScheduleTypeEnum, DayOfWeekEnum, ScheduleVersionEnum
from storops.unity.resource.snap_schedule import UnitySnapSchedule, \
    UnitySnapScheduleRule, UnitySnapScheduleList
from storops_test.unity.rest_mock import t_rest, patch_rest

__author__ = 'Ryan Liang'


@ddt.ddt
class UnitySnapScheduleRuleTest(TestCase):
    @ddt.data(
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': None, 'days_of_week': None, 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'expected':
             {'type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
              'hours': None, 'daysOfWeek': None, 'daysOfMonth': None,
              'interval': None, 'isAutoDelete': None, 'retentionTime': None,
              'accessType': None, }},
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': 0,
         'hours': [8], 'days_of_week': [DayOfWeekEnum.MONDAY],
         'days_of_month': [10],
         'interval': 15, 'is_auto_delete': True, 'retention_time': 3600,
         'access_type': FilesystemSnapAccessTypeEnum.CHECKPOINT,
         'expected':
             {'type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': 0,
              'hours': [8], 'daysOfWeek': [DayOfWeekEnum.MONDAY],
              'daysOfMonth': [10],
              'interval': 15, 'isAutoDelete': True, 'retentionTime': 3600,
              'accessType': FilesystemSnapAccessTypeEnum.CHECKPOINT, }},
    )
    @ddt.unpack
    def test_to_embedded(self, schedule_type, minute, hours, days_of_week,
                         days_of_month, interval, is_auto_delete,
                         retention_time, access_type, expected):
        assert_that(
            UnitySnapScheduleRule.to_embedded(schedule_type, minute=minute,
                                              hours=hours,
                                              days_of_week=days_of_week,
                                              days_of_month=days_of_month,
                                              interval=interval,
                                              is_auto_delete=is_auto_delete,
                                              retention_time=retention_time,
                                              access_type=access_type),
            equal_to(expected))

    @ddt.data(
        {'schedule_type': None, 'minute': None,
         'hours': None, 'days_of_week': None, 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': 'None is not allowed here for ScheduleTypeEnum'},
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': -1,
         'hours': None, 'days_of_week': None, 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg':
             'minute in UnitySnapScheduleRule should be \[0, 59\].'},  # noqa
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': 60,
         'hours': None, 'days_of_week': None, 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg':
             'minute in UnitySnapScheduleRule should be \[0, 59\].'},  # noqa
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': 10, 'days_of_week': None, 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': 'hours in UnitySnapScheduleRule should be a list '
                      'and not an empty list.'},
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': [], 'days_of_week': None, 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': 'hours in UnitySnapScheduleRule should be a list '
                      'and not an empty list.'},
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': [-1, 10], 'days_of_week': None, 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': 'each value of hours in UnitySnapScheduleRule should be '
                      '\[0, 23\].'},  # noqa
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': [24, 10], 'days_of_week': None, 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': 'each value of hours in UnitySnapScheduleRule should be '
                      '\[0, 23\].'},  # noqa
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': None, 'days_of_week': 1, 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': 'days_of_week in UnitySnapScheduleRule should be a list '
                      'and not an empty list.'},
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': None, 'days_of_week': [], 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': 'days_of_week in UnitySnapScheduleRule should be a list '
                      'and not an empty list.'},
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': None, 'days_of_week': [0, 3], 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': '0 is not an instance of DayOfWeekEnum.'},
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': None, 'days_of_week': [8, 3], 'days_of_month': None,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': '8 is not an instance of DayOfWeekEnum.'},
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': None, 'days_of_week': None, 'days_of_month': 1,
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': 'days_of_month in UnitySnapScheduleRule should be a '
                      'list and not an empty list.'},
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': None, 'days_of_week': None, 'days_of_month': [],
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': 'days_of_month in UnitySnapScheduleRule should be a '
                      'list and not an empty list.'},
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': None, 'days_of_week': None, 'days_of_month': [0, 3],
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': 'each value of days_of_month in UnitySnapScheduleRule '
                      'should be \[1, 31\].'},  # noqa
        {'schedule_type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': None,
         'hours': None, 'days_of_week': None, 'days_of_month': [32, 3],
         'interval': None, 'is_auto_delete': None, 'retention_time': None,
         'access_type': None,
         'error_msg': 'each value of days_of_month in UnitySnapScheduleRule '
                      'should be \[1, 31\].'},  # noqa
    )
    @ddt.unpack
    def test_to_embedded_value_error(self, schedule_type, minute, hours,
                                     days_of_week, days_of_month, interval,
                                     is_auto_delete, retention_time,
                                     access_type, error_msg):
        assert_that(
            calling(UnitySnapScheduleRule.to_embedded).with_args(
                schedule_type, minute=minute, hours=hours,
                days_of_week=days_of_week, days_of_month=days_of_month,
                interval=interval, is_auto_delete=is_auto_delete,
                retention_time=retention_time, access_type=access_type),
            raises(ValueError, error_msg))

    @ddt.data(
        {'hour_interval': 3, 'minute': None,
         'is_auto_delete': None, 'retention_time': None, 'access_type': None,
         'expected':
             {'type': ScheduleTypeEnum.N_HOURS_AT_MM, 'minute': 0,
              'hours': None, 'daysOfWeek': None,
              'daysOfMonth': None,
              'interval': 3, 'isAutoDelete': None, 'retentionTime': None,
              'accessType': None}},
        {'hour_interval': 3, 'minute': 30,
         'is_auto_delete': True, 'retention_time': 3600,
         'access_type': FilesystemSnapAccessTypeEnum.CHECKPOINT,
         'expected':
             {'type': ScheduleTypeEnum.N_HOURS_AT_MM, 'minute': 30,
              'hours': None, 'daysOfWeek': None,
              'daysOfMonth': None,
              'interval': 3, 'isAutoDelete': True, 'retentionTime': 3600,
              'accessType': FilesystemSnapAccessTypeEnum.CHECKPOINT}},
    )
    def test_every_n_hours(self, kwargs):
        hour_interval = kwargs.pop('hour_interval')
        expected = kwargs.pop('expected')
        kwargs = {key: val for key, val in kwargs.items() if val is not None}
        assert_that(UnitySnapScheduleRule.every_n_hours(hour_interval,
                                                        **kwargs),
                    equal_to(expected))

    @ddt.data(
        {'hours': [0, 12], 'minute': None,
         'is_auto_delete': None, 'retention_time': None, 'access_type': None,
         'expected':
             {'type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': 0,
              'hours': [0, 12], 'daysOfWeek': None,
              'daysOfMonth': None,
              'interval': 1, 'isAutoDelete': None, 'retentionTime': None,
              'accessType': None}},
        {'hours': [0, 12], 'minute': 30,
         'is_auto_delete': True, 'retention_time': 3600,
         'access_type': FilesystemSnapAccessTypeEnum.CHECKPOINT,
         'expected':
             {'type': ScheduleTypeEnum.DAY_AT_HHMM, 'minute': 30,
              'hours': [0, 12], 'daysOfWeek': None,
              'daysOfMonth': None,
              'interval': 1, 'isAutoDelete': True, 'retentionTime': 3600,
              'accessType': FilesystemSnapAccessTypeEnum.CHECKPOINT}},
    )
    def test_every_day(self, kwargs):
        hours = kwargs.pop('hours')
        expected = kwargs.pop('expected')
        kwargs = {key: val for key, val in kwargs.items() if val is not None}
        assert_that(UnitySnapScheduleRule.every_day(hours,
                                                    **kwargs),
                    equal_to(expected))

    @ddt.data(
        {'hours': 12, 'minute': None,
         'is_auto_delete': None, 'retention_time': None, 'access_type': None,
         'error_msg': 'hours of UnitySnapScheduleRule every_day rule should '
                      'be a list.'},
        {'hours': [], 'minute': None,
         'is_auto_delete': None, 'retention_time': None, 'access_type': None,
         'error_msg': 'hours of UnitySnapScheduleRule every_day rule cannot '
                      'be empty list.'}
    )
    def test_every_day_value_error(self, kwargs):
        hours = kwargs.pop('hours')
        error_msg = kwargs.pop('error_msg')
        kwargs = {key: val for key, val in kwargs.items() if val is not None}
        assert_that(
            calling(UnitySnapScheduleRule.every_day).with_args(hours,
                                                               **kwargs),
            raises(ValueError, error_msg))

    @ddt.data(
        {'day_interval': 3, 'hour': 0, 'minute': None,
         'is_auto_delete': None, 'retention_time': None, 'access_type': None,
         'expected':
             {'type': ScheduleTypeEnum.N_DAYS_AT_HHMM, 'minute': 0,
              'hours': [0], 'daysOfWeek': None,
              'daysOfMonth': None,
              'interval': 3, 'isAutoDelete': None, 'retentionTime': None,
              'accessType': None}},
        {'day_interval': 3, 'hour': 0, 'minute': 30,
         'is_auto_delete': True, 'retention_time': 3600,
         'access_type': FilesystemSnapAccessTypeEnum.CHECKPOINT,
         'expected':
             {'type': ScheduleTypeEnum.N_DAYS_AT_HHMM, 'minute': 30,
              'hours': [0], 'daysOfWeek': None,
              'daysOfMonth': None,
              'interval': 3, 'isAutoDelete': True, 'retentionTime': 3600,
              'accessType': FilesystemSnapAccessTypeEnum.CHECKPOINT}},
    )
    def test_every_n_days(self, kwargs):
        day_interval = kwargs.pop('day_interval')
        expected = kwargs.pop('expected')
        kwargs = {key: val for key, val in kwargs.items() if val is not None}
        assert_that(UnitySnapScheduleRule.every_n_days(day_interval,
                                                       **kwargs),
                    equal_to(expected))

    @ddt.data(
        {'days_of_week': [DayOfWeekEnum.TUESDAY, DayOfWeekEnum.THURSDAY],
         'hour': None, 'minute': None,
         'is_auto_delete': None, 'retention_time': None, 'access_type': None,
         'expected':
             {'type': ScheduleTypeEnum.SELDAYS_AT_HHMM, 'minute': 0,
              'hours': [0],
              'daysOfWeek': [DayOfWeekEnum.TUESDAY, DayOfWeekEnum.THURSDAY],
              'daysOfMonth': None,
              'interval': None, 'isAutoDelete': None, 'retentionTime': None,
              'accessType': None}},
        {'days_of_week': [DayOfWeekEnum.TUESDAY, DayOfWeekEnum.THURSDAY],
         'hour': 12, 'minute': 30,
         'is_auto_delete': True, 'retention_time': 3600,
         'access_type': FilesystemSnapAccessTypeEnum.CHECKPOINT,
         'expected':
             {'type': ScheduleTypeEnum.SELDAYS_AT_HHMM, 'minute': 30,
              'hours': [12],
              'daysOfWeek': [DayOfWeekEnum.TUESDAY, DayOfWeekEnum.THURSDAY],
              'daysOfMonth': None,
              'interval': None, 'isAutoDelete': True, 'retentionTime': 3600,
              'accessType': FilesystemSnapAccessTypeEnum.CHECKPOINT}},
    )
    def test_every_week(self, kwargs):
        days_of_week = kwargs.pop('days_of_week')
        expected = kwargs.pop('expected')
        kwargs = {key: val for key, val in kwargs.items() if val is not None}
        assert_that(UnitySnapScheduleRule.every_week(days_of_week,
                                                     **kwargs),
                    equal_to(expected))

    @ddt.data(
        {'days_of_week': 2, 'hour': None, 'minute': None,
         'is_auto_delete': None, 'retention_time': None, 'access_type': None,
         'error_msg': 'days_of_week of UnitySnapScheduleRule every_week rule '
                      'should be a list.'},
        {'days_of_week': [], 'hour': None, 'minute': None,
         'is_auto_delete': None, 'retention_time': None, 'access_type': None,
         'error_msg': 'days_of_week of UnitySnapScheduleRule every_week rule '
                      'cannot be empty list.'}
    )
    def test_every_week_value_error(self, kwargs):
        days_of_week = kwargs.pop('days_of_week')
        error_msg = kwargs.pop('error_msg')
        kwargs = {key: val for key, val in kwargs.items() if val is not None}
        assert_that(
            calling(UnitySnapScheduleRule.every_week).with_args(days_of_week,
                                                                **kwargs),
            raises(ValueError, error_msg))

    @ddt.data(
        {'days_of_month': [10, 20], 'hour': None, 'minute': None,
         'is_auto_delete': None, 'retention_time': None, 'access_type': None,
         'expected':
             {'type': ScheduleTypeEnum.NTH_DAYOFMONTH_AT_HHMM,
              'minute': 0, 'hours': [0],
              'daysOfWeek': None,
              'daysOfMonth': [10, 20],
              'interval': None, 'isAutoDelete': None, 'retentionTime': None,
              'accessType': None}},
        {'days_of_month': [10, 20], 'hour': 12, 'minute': 30,
         'is_auto_delete': True, 'retention_time': 3600,
         'access_type': FilesystemSnapAccessTypeEnum.CHECKPOINT,
         'expected':
             {'type': ScheduleTypeEnum.NTH_DAYOFMONTH_AT_HHMM,
              'minute': 30, 'hours': [12],
              'daysOfWeek': None,
              'daysOfMonth': [10, 20],
              'interval': None, 'isAutoDelete': True, 'retentionTime': 3600,
              'accessType': FilesystemSnapAccessTypeEnum.CHECKPOINT}},
    )
    def test_every_month(self, kwargs):
        days_of_month = kwargs.pop('days_of_month')
        expected = kwargs.pop('expected')
        kwargs = {key: val for key, val in kwargs.items() if val is not None}
        assert_that(UnitySnapScheduleRule.every_month(days_of_month,
                                                      **kwargs),
                    equal_to(expected))

    @ddt.data(
        {'days_of_month': 10, 'hour': None, 'minute': None,
         'is_auto_delete': None, 'retention_time': None, 'access_type': None,
         'error_msg': 'days_of_month of UnitySnapScheduleRule every_month '
                      'rule should be a list.'},
        {'days_of_month': [], 'hour': None, 'minute': None,
         'is_auto_delete': None, 'retention_time': None, 'access_type': None,
         'error_msg': 'days_of_month of UnitySnapScheduleRule every_month '
                      'rule cannot be empty list.'}
    )
    def test_every_month_value_error(self, kwargs):
        days_of_month = kwargs.pop('days_of_month')
        error_msg = kwargs.pop('error_msg')
        kwargs = {key: val for key, val in kwargs.items() if val is not None}
        assert_that(
            calling(UnitySnapScheduleRule.every_month).with_args(days_of_month,
                                                                 **kwargs),
            raises(ValueError, error_msg))


@ddt.ddt
class UnitySnapScheduleTest(TestCase):
    @patch_rest
    def test_properties(self):
        schedule = UnitySnapSchedule(_id='snapSch_1', cli=t_rest())
        assert_that(schedule.existed, equal_to(True))
        assert_that(schedule.version, equal_to(ScheduleVersionEnum.LEGACY))
        assert_that(schedule.name, equal_to('Snapshot Schedule 1'))
        assert_that(schedule.is_default, equal_to(False))
        assert_that(schedule.is_modified, equal_to(False))
        assert_that(str(schedule.modification_time),
                    equal_to('2020-03-17 07:37:56.695000+00:00'))
        assert_that(len(schedule.rules), equal_to(1))
        assert_that(schedule.rules[0].type,
                    equal_to(ScheduleTypeEnum.DAY_AT_HHMM))
        assert_that(len(schedule.rules[0].days_of_week), equal_to(0))
        assert_that(schedule.rules[0].access_type,
                    equal_to(FilesystemSnapAccessTypeEnum.CHECKPOINT))
        assert_that(schedule.rules[0].minute, equal_to(0))
        assert_that(schedule.rules[0].hours, equal_to([4]))
        assert_that(schedule.rules[0].interval, equal_to(1))
        assert_that(schedule.rules[0].is_auto_delete, equal_to(True))
        assert_that(schedule.rules[0].retention_time, equal_to(0))
        assert_that(schedule.is_sync_replicated, equal_to(False))
        assert_that(len(schedule.luns), equal_to(1))
        assert_that(schedule.luns[0].get_id(), equal_to('sv_2'))

    @patch_rest
    def test_get_all(self):
        schedules = UnitySnapScheduleList(cli=t_rest())
        assert_that(schedules[0].existed, equal_to(True))
        assert_that(len(schedules), equal_to(3))

    @ddt.data(
        {'is_sync_replicated': None, 'skip_sync_to_remote_system': None,
         'expected': 'snapSch_4'},
        {'is_sync_replicated': True, 'skip_sync_to_remote_system': True,
         'expected': 'snapSch_5'},
    )
    @patch_rest
    def test_create(self, kwargs):
        expected = kwargs.pop('expected')
        kwargs = {key: val for key, val in kwargs.items() if val is not None}
        schedule = UnitySnapSchedule.create(
            t_rest(), 'snap-schedule', [UnitySnapScheduleRule.every_day([12])],
            **kwargs
        )
        assert_that(schedule.get_id(), equal_to(expected))

    @ddt.data(
        {'add_rules': [UnitySnapScheduleRule.every_month([10])],
         'remove_rule_ids': None,
         'skip_sync_to_remote_system': None},
        {'add_rules': None,
         'remove_rule_ids': ['SchedRule_4'],
         'skip_sync_to_remote_system': None},
        {'add_rules': None,
         'remove_rule_ids': None,
         'skip_sync_to_remote_system': True},
    )
    @patch_rest
    def test_modify(self, kwargs):
        schedule = UnitySnapSchedule(_id='snapSch_4', cli=t_rest())
        kwargs = {key: val for key, val in kwargs.items() if val is not None}
        resp = schedule.modify(**kwargs)
        assert_that(resp.is_ok(), equal_to(True))

    @patch_rest
    def test_delete(self):
        schedule = UnitySnapSchedule(_id='snapSch_4', cli=t_rest())
        resp = schedule.delete()
        assert_that(resp.is_ok(), equal_to(True))
